from pcs.common import reports
from pcs.common.types import StringCollection

from . import const


class ResourceAgentError(Exception):
    def __init__(self, agent_name: str):
        super().__init__()
        self.agent_name = agent_name


class AgentNameGuessFoundMoreThanOne(ResourceAgentError):
    def __init__(self, searched_name: str, names_found: StringCollection):
        super().__init__(searched_name)
        self.names_found = names_found

    @property
    def searched_name(self) -> str:
        return self.agent_name


class AgentNameGuessFoundNone(ResourceAgentError):
    @property
    def searched_name(self) -> str:
        return self.agent_name


class InvalidResourceAgentName(ResourceAgentError):
    pass


class UnableToGetAgentMetadata(ResourceAgentError):
    def __init__(self, agent_name: str, message: str):
        super().__init__(agent_name)
        self.message = message


class UnsupportedOcfVersion(ResourceAgentError):
    def __init__(self, agent_name: str, ocf_version: str):
        super().__init__(agent_name)
        self.ocf_version = ocf_version


def resource_agent_error_to_report_item(
    e: ResourceAgentError,
    severity: reports.ReportItemSeverity = reports.ReportItemSeverity.error(),
    is_stonith: bool = False,
) -> reports.ReportItem:
    """
    Transform a ResourceAgentError to a ReportItem
    """
    message: reports.item.ReportItemMessage = (
        reports.messages.AgentGenericError(e.agent_name)
    )
    if isinstance(e, AgentNameGuessFoundMoreThanOne):
        message = reports.messages.AgentNameGuessFoundMoreThanOne(
            e.searched_name, sorted(e.names_found)
        )
    elif isinstance(e, AgentNameGuessFoundNone):
        message = reports.messages.AgentNameGuessFoundNone(e.searched_name)
    elif isinstance(e, InvalidResourceAgentName):
        if is_stonith:
            message = reports.messages.InvalidStonithAgentName(e.agent_name)
        else:
            message = reports.messages.InvalidResourceAgentName(e.agent_name)
    elif isinstance(e, UnableToGetAgentMetadata):
        message = reports.messages.UnableToGetAgentMetadata(
            e.agent_name, e.message
        )
    elif isinstance(e, UnsupportedOcfVersion):
        message = reports.messages.AgentImplementsUnsupportedOcfVersion(
            e.agent_name,
            e.ocf_version,
            sorted(const.SUPPORTED_OCF_VERSIONS),
        )
    return reports.ReportItem(severity, message)



import sys
from typing import Optional

from pcs.cli.common.tools import print_to_stderr
from pcs.common.reports import (
    ReportItemList,
    ReportItemSeverity,
)
from pcs.common.reports.dto import ReportItemContextDto

from .messages import report_item_msg_from_dto


def warn(message: str) -> None:
    print_to_stderr(f"Warning: {message}")


def deprecation_warning(message: str) -> None:
    print_to_stderr(f"Deprecation Warning: {message}")


def error(message: str) -> SystemExit:
    print_to_stderr(f"Error: {message}")
    return SystemExit(1)


def add_context_to_message(
    msg: str, context: Optional[ReportItemContextDto]
) -> str:
    if context:
        msg = f"{context.node}: {msg}"
    return msg


def process_library_reports(report_item_list: ReportItemList) -> None:
    if not report_item_list:
        raise error("Errors have occurred, therefore pcs is unable to continue")

    critical_error = False
    for report_item in report_item_list:
        report_dto = report_item.to_dto()
        cli_report_msg = report_item_msg_from_dto(report_dto.message)
        msg = add_context_to_message(cli_report_msg.message, report_dto.context)
        severity = report_dto.severity.level

        if severity == ReportItemSeverity.WARNING:
            warn(msg)
            continue

        if severity == ReportItemSeverity.DEPRECATION:
            deprecation_warning(msg)
            continue

        if severity != ReportItemSeverity.ERROR:
            print_to_stderr(msg)
            continue

        error(
            add_context_to_message(
                cli_report_msg.get_message_with_force_text(
                    report_item.severity.force_code
                ),
                report_dto.context,
            )
        )
        critical_error = True

    if critical_error:
        sys.exit(1)



from pcs import (
    alert,
    usage,
)
from pcs.cli.common.routing import create_router

alert_cmd = create_router(
    {
        "help": lambda lib, argv, modifiers: print(usage.alert(argv)),
        "create": alert.alert_add,
        "update": alert.alert_update,
        "delete": alert.alert_remove,
        "remove": alert.alert_remove,
        "config": alert.print_alert_config,
        # TODO remove, deprecated command
        # replaced with 'config'
        "show": alert.print_alert_show,
        "recipient": create_router(
            {
                "help": lambda lib, argv, modifiers: print(
                    usage.alert(["recipient"])
                ),
                "add": alert.recipient_add,
                "update": alert.recipient_update,
                "delete": alert.recipient_remove,
                "remove": alert.recipient_remove,
            },
            ["alert", "recipient"],
        ),
        "get_all_alerts": alert.print_alerts_in_json,
    },
    ["alert"],
    default_cmd="config",
)



import os

from pcs import settings
from pcs.common import (
    file_type_codes,
    reports,
)
from pcs.common.reports.item import ReportItem
from pcs.lib.file.instance import FileInstance


def get_all_configs_file_names():
    """
    Get a list of all files ending with '.conf' from booth configuration dir.
    """
    if not os.path.isdir(settings.booth_config_dir):
        return []
    return [
        file_name
        for file_name in os.listdir(settings.booth_config_dir)
        if file_name.endswith(".conf")
        and len(file_name) > len(".conf")
        and os.path.isfile(os.path.join(settings.booth_config_dir, file_name))
    ]


def get_authfile_name_and_data(booth_conf_facade):
    """
    Get booth auth filename, content and reports based on booth config facade

    pcs.lib.booth.config_facade.ConfigFacade booth_conf_facade -- booth config
    """
    authfile_name = None
    authfile_data = None
    report_list = []

    authfile_path = booth_conf_facade.get_authfile()
    if authfile_path:
        authfile_dir, authfile_name = os.path.split(authfile_path)
        if (authfile_dir == settings.booth_config_dir) and authfile_name:
            authfile_data = FileInstance.for_booth_key(authfile_name).read_raw()
        else:
            authfile_name = None
            report_list.append(
                ReportItem.warning(
                    reports.messages.BoothUnsupportedFileLocation(
                        authfile_path,
                        settings.booth_config_dir,
                        file_type_codes.BOOTH_KEY,
                    )
                )
            )

    return authfile_name, authfile_data, report_list



from textwrap import dedent
from unittest import TestCase

from lxml import etree

from pcs_test.tools.assertions import AssertPcsMixinOld as AssertPcsMixin
from pcs_test.tools.cib import (
    get_assert_pcs_effect_mixin_old as get_assert_pcs_effect_mixin,
)
from pcs_test.tools.misc import get_test_resource as rc
from pcs_test.tools.misc import (
    get_tmp_file,
    skip_unless_crm_rule,
    skip_unless_pacemaker_supports_rsc_and_op_rules,
    write_data_to_tmpfile,
    write_file_to_tmpfile,
)
from pcs_test.tools.pcs_runner import PcsRunnerOld as PcsRunner
from pcs_test.tools.xml import XmlManipulation

empty_cib = rc("cib-empty.xml")
empty_cib_rules = rc("cib-empty-3.4.xml")


class TestDefaultsMixin:
    def setUp(self):
        # pylint: disable=invalid-name
        self.temp_cib = get_tmp_file("tier1_cib_options")
        self.pcs_runner = PcsRunner(self.temp_cib.name)

    def tearDown(self):
        # pylint: disable=invalid-name
        self.temp_cib.close()


class DefaultsConfigMixin(TestDefaultsMixin, AssertPcsMixin):
    cli_command = []
    prefix = ""

    def test_success(self):
        xml_rsc = """
            <rsc_defaults>
                <meta_attributes id="rsc-set1" score="10">
                    <nvpair id="rsc-set1-nv1" name="name1" value="rsc1"/>
                    <nvpair id="rsc-set1-nv2" name="name2" value="rsc2"/>
                </meta_attributes>
                <meta_attributes id="rsc-setA">
                    <nvpair id="rsc-setA-nv1" name="name1" value="rscA"/>
                    <nvpair id="rsc-setA-nv2" name="name2" value="rscB"/>
                </meta_attributes>
            </rsc_defaults>
        """
        xml_op = """
            <op_defaults>
                <meta_attributes id="op-set1" score="10">
                    <nvpair id="op-set1-nv1" name="name1" value="op1"/>
                    <nvpair id="op-set1-nv2" name="name2" value="op2"/>
                </meta_attributes>
                <meta_attributes id="op-setA">
                    <nvpair id="op-setA-nv1" name="name1" value="opA"/>
                    <nvpair id="op-setA-nv2" name="name2" value="opB"/>
                </meta_attributes>
            </op_defaults>
        """
        xml_manip = XmlManipulation.from_file(empty_cib)
        xml_manip.append_to_first_tag_name("configuration", xml_rsc, xml_op)
        write_data_to_tmpfile(str(xml_manip), self.temp_cib)

        self.assert_pcs_success(
            self.cli_command,
            stdout_full=dedent(
                f"""\
                Meta Attrs: {self.prefix}-set1 score=10
                  name1={self.prefix}1
                  name2={self.prefix}2
                Meta Attrs: {self.prefix}-setA
                  name1={self.prefix}A
                  name2={self.prefix}B
            """
            ),
        )

    def test_success_rule(self):
        xml_template = """<{tag}_defaults>
            <meta_attributes id="{tag}-set1">
                <rule id="{tag}-set1-rule" boolean-op="and" score="INFINITY">
                    <rule id="{tag}-set1-rule-rule" boolean-op="or" score="0">
                        <expression id="{tag}-set1-rule-rule-expr"
                            operation="defined" attribute="attr1"
                        />
                        <expression id="{tag}-set1-rule-rule-expr-1"
                            attribute="attr2" operation="gte"
                            type="number" value="12"
                        />
                        <expression id="{tag}-set1-rule-rule-expr-2"
                            attribute="attr3" operation="lt"
                            type="version" value="3.2.1"
                        />
                        <expression id="{tag}-set1-rule-rule-expr-3"
                            attribute="attr4" operation="ne"
                            type="string" value="test"
                        />
                        <expression id="{tag}-set1-rule-rule-expr-4"
                            attribute="attr5" operation="lt" value="3"
                        />
                    </rule>
                    <rule id="{tag}-set1-rule-rule-1" boolean-op="or" score="0">
                        <date_expression id="{tag}-set1-rule-rule-1-expr"
                            operation="gt" start="2018-05-17T13:28:19"
                        />
                        <date_expression id="{tag}-set1-rule-rule-1-expr-1"
                            operation="in_range"
                            start="2019-01-01" end="2019-03-15"
                        />
                        <date_expression id="{tag}-set1-rule-rule-1-expr-2"
                             operation="in_range" start="2019-05-01"
                        >
                            <duration id="{tag}-set1-rule-rule-1-expr-2-duration"
                                months="2"
                            />
                        </date_expression>
                        <date_expression id="{tag}-set1-rule-rule-1-expr-3"
                            operation="date_spec"
                        >
                            <date_spec id="{tag}-set1-rule-rule-1-expr-3-datespec"
                                months="7-8" weekdays="6-7" years="2019"
                            />
                        </date_expression>
                        <date_expression id="{tag}-set1-rule-rule-1-expr-4"
                            operation="in_range" end="2019-12-15"
                        />
                    </rule>
                </rule>
                <nvpair id="{tag}-set1-nam1" name="nam1" value="val1"/>
                <nvpair id="{tag}-set1-nam2" name="nam2" value="val2"/>
            </meta_attributes>
        </{tag}_defaults>"""
        xml_rsc = xml_template.format(tag="rsc")
        xml_op = xml_template.format(tag="op")
        xml_manip = XmlManipulation.from_file(empty_cib)
        xml_manip.append_to_first_tag_name("configuration", xml_rsc, xml_op)
        write_data_to_tmpfile(str(xml_manip), self.temp_cib)

        self.assert_pcs_success(
            self.cli_command,
            stdout_full=dedent(
                f"""\
                Meta Attrs: {self.prefix}-set1
                  nam1=val1
                  nam2=val2
                  Rule: boolean-op=and score=INFINITY
                    Rule: boolean-op=or score=0
                      Expression: defined attr1
                      Expression: attr2 gte number 12
                      Expression: attr3 lt version 3.2.1
                      Expression: attr4 ne string test
                      Expression: attr5 lt 3
                    Rule: boolean-op=or score=0
                      Expression: date gt 2018-05-17T13:28:19
                      Expression: date in_range 2019-01-01 to 2019-03-15
                      Expression: date in_range 2019-05-01 to duration
                        Duration: months=2
                      Expression:
                        Date Spec: months=7-8 weekdays=6-7 years=2019
                      Expression: date in_range to 2019-12-15
            """
            ),
        )

    xml_expired_template = """<{tag}_defaults>
        <meta_attributes id="{tag}-set1">
            <rule id="{tag}-set1-rule" boolean-op="and" score="INFINITY">
                <date_expression id="{tag}-set1-rule-expr"
                    operation="gt" start="3000-01-01"
                />
            </rule>
            <nvpair id="{tag}-set1-name" name="name1" value="value1"/>
        </meta_attributes>
        <meta_attributes id="{tag}-set2">
            <rule id="{tag}-set2-rule" boolean-op="and" score="INFINITY">
                <date_expression id="{tag}-set2-rule-expr"
                    operation="lt" end="1000-01-01"
                />
            </rule>
            <nvpair id="{tag}-set2-name" name="name2" value="value2"/>
        </meta_attributes>
        <meta_attributes id="{tag}-set3">
            <rule id="{tag}-set3-rule" boolean-op="and" score="INFINITY">
                <date_expression id="{tag}-set3-rule-expr"
                    operation="in_range" start="1000-01-01" end="3000-01-01"
                />
            </rule>
            <nvpair id="{tag}-set3-name" name="name3" value="value3"/>
        </meta_attributes>
    </{tag}_defaults>"""

    @skip_unless_crm_rule()
    def test_success_rule_expired(self):
        xml_rsc = self.xml_expired_template.format(tag="rsc")
        xml_op = self.xml_expired_template.format(tag="op")
        xml_manip = XmlManipulation.from_file(empty_cib)
        xml_manip.append_to_first_tag_name("configuration", xml_rsc, xml_op)
        write_data_to_tmpfile(str(xml_manip), self.temp_cib)

        self.assert_pcs_success(
            self.cli_command,
            stdout_full=dedent(
                f"""\
                Meta Attrs (not yet in effect): {self.prefix}-set1
                  name1=value1
                  Rule (not yet in effect): boolean-op=and score=INFINITY
                    Expression: date gt 3000-01-01
                Meta Attrs: {self.prefix}-set3
                  name3=value3
                  Rule: boolean-op=and score=INFINITY
                    Expression: date in_range 1000-01-01 to 3000-01-01
            """
            ),
        )

    @skip_unless_crm_rule()
    def test_success_rule_expired_all(self):
        xml_rsc = self.xml_expired_template.format(tag="rsc")
        xml_op = self.xml_expired_template.format(tag="op")
        xml_manip = XmlManipulation.from_file(empty_cib)
        xml_manip.append_to_first_tag_name("configuration", xml_rsc, xml_op)
        write_data_to_tmpfile(str(xml_manip), self.temp_cib)

        self.assert_pcs_success(
            self.cli_command + ["--all"],
            stdout_full=dedent(
                f"""\
                Meta Attrs (not yet in effect): {self.prefix}-set1
                  name1=value1
                  Rule (not yet in effect): boolean-op=and score=INFINITY
                    Expression: date gt 3000-01-01
                Meta Attrs (expired): {self.prefix}-set2
                  name2=value2
                  Rule (expired): boolean-op=and score=INFINITY
                    Expression: date lt 1000-01-01
                Meta Attrs: {self.prefix}-set3
                  name3=value3
                  Rule: boolean-op=and score=INFINITY
                    Expression: date in_range 1000-01-01 to 3000-01-01
            """
            ),
        )


class RscDefaultsConfig(
    DefaultsConfigMixin,
    TestCase,
):
    cli_command = ["resource", "defaults"]
    prefix = "rsc"

    @skip_unless_pacemaker_supports_rsc_and_op_rules()
    def test_success_rules_rsc_op(self):
        xml = """
            <rsc_defaults>
                <meta_attributes id="X">
                    <rule id="X-rule" boolean-op="and" score="INFINITY">
                        <rsc_expression id="X-rule-rsc-Dummy" type="Dummy"/>
                    </rule>
                    <nvpair id="X-nam1" name="nam1" value="val1"/>
                </meta_attributes>
            </rsc_defaults>
        """
        xml_manip = XmlManipulation.from_file(empty_cib_rules)
        xml_manip.append_to_first_tag_name("configuration", xml)
        write_data_to_tmpfile(str(xml_manip), self.temp_cib)

        self.assert_pcs_success(
            self.cli_command,
            stdout_full=dedent(
                """\
                Meta Attrs: X
                  nam1=val1
                  Rule: boolean-op=and score=INFINITY
                    Expression: resource ::Dummy
            """
            ),
        )


class OpDefaultsConfig(
    DefaultsConfigMixin,
    TestCase,
):
    cli_command = ["resource", "op", "defaults"]
    prefix = "op"

    @skip_unless_pacemaker_supports_rsc_and_op_rules()
    def test_success_rules_rsc_op(self):
        xml = """
            <op_defaults>
                <meta_attributes id="X">
                    <rule id="X-rule" boolean-op="and" score="INFINITY">
                        <rsc_expression id="X-rule-rsc-Dummy" type="Dummy"/>
                        <op_expression id="X-rule-op-monitor" name="monitor"/>
                    </rule>
                    <nvpair id="X-nam1" name="nam1" value="val1"/>
                </meta_attributes>
            </op_defaults>
        """
        xml_manip = XmlManipulation.from_file(empty_cib_rules)
        xml_manip.append_to_first_tag_name("configuration", xml)
        write_data_to_tmpfile(str(xml_manip), self.temp_cib)

        self.assert_pcs_success(
            self.cli_command,
            stdout_full=dedent(
                """\
                Meta Attrs: X
                  nam1=val1
                  Rule: boolean-op=and score=INFINITY
                    Expression: resource ::Dummy
                    Expression: op monitor
            """
            ),
        )


class DefaultsSetCreateMixin(TestDefaultsMixin, AssertPcsMixin):
    cli_command = []
    cib_tag = ""

    def setUp(self):
        super().setUp()
        write_file_to_tmpfile(empty_cib, self.temp_cib)

    def test_no_args(self):
        self.assert_effect(
            self.cli_command + ["set", "create"],
            dedent(
                f"""\
                <{self.cib_tag}>
                    <meta_attributes id="{self.cib_tag}-meta_attributes"/>
                </{self.cib_tag}>
            """
            ),
            output=(
                "Warning: Defaults do not apply to resources which override "
                "them with their own defined values\n"
            ),
        )

    def test_success(self):
        self.assert_effect(
            self.cli_command
            + "set create id=mine score=10 meta nam1=val1 nam2=val2 --force".split(),
            dedent(
                f"""\
                <{self.cib_tag}>
                    <meta_attributes id="mine" score="10">
                        <nvpair id="mine-nam1" name="nam1" value="val1"/>
                        <nvpair id="mine-nam2" name="nam2" value="val2"/>
                    </meta_attributes>
                </{self.cib_tag}>
            """
            ),
            output=(
                "Warning: Defaults do not apply to resources which override "
                "them with their own defined values\n"
            ),
        )

    def test_success_rule(self):
        self.assert_effect(
            self.cli_command
            + (
                "-- set create id=mine score=10 meta nam1=val1 nam2=val2 "
                "rule (date gt 2018-05-17T13:28:19 or "
                "date in_range 2019-01-01 to 2019-03-15 or "
                "date in_range 2019-05-01 to duration months=2 or "
                "date-spec years=2019 months=7-8 weekdays=6-7 or "
                "date in_range to 2019-12-15)"
            ).split(),
            dedent(
                f"""\
                <{self.cib_tag}>
                    <meta_attributes id="mine" score="10">
                        <rule id="mine-rule" boolean-op="or" score="INFINITY">
                            <date_expression id="mine-rule-expr"
                                operation="gt" start="2018-05-17T13:28:19"
                            />
                            <date_expression id="mine-rule-expr-1"
                                operation="in_range"
                                start="2019-01-01" end="2019-03-15"
                            />
                            <date_expression id="mine-rule-expr-2"
                                operation="in_range" start="2019-05-01"
                            >
                                <duration id="mine-rule-expr-2-duration"
                                    months="2"
                                />
                            </date_expression>
                            <date_expression id="mine-rule-expr-3"
                                operation="date_spec"
                            >
                                <date_spec
                                    id="mine-rule-expr-3-datespec"
                                    months="7-8" weekdays="6-7" years="2019"
                                />
                            </date_expression>
                            <date_expression id="mine-rule-expr-4"
                                  operation="in_range" end="2019-12-15"
                            />
                        </rule>
                        <nvpair id="mine-nam1" name="nam1" value="val1"/>
                        <nvpair id="mine-nam2" name="nam2" value="val2"/>
                    </meta_attributes>
                </{self.cib_tag}>
            """
            ),
            output=(
                "Warning: Defaults do not apply to resources which override "
                "them with their own defined values\n"
            ),
        )

    def test_rule_error_messages(self):
        self.assert_pcs_fail(
            self.cli_command
            + (
                "set create id=mine score=10 meta nam1=val1 nam2=val2 "
                "rule (date gt 2018-05-1X or "
                "date in_range 2019-03-05 to 2019-01-11 or "
                "date in_range 2019-05-0X to duration months=2 months=3a x=y or "
                "date-spec years=2019 months=7-X weekdays=7-6 years=202a x=y)"
            ).split(),
            (
                "Error: '2018-05-1X' is not a valid date value, use ISO 8601 date\n"
                "Error: Since '2019-03-05' is not sooner than until '2019-01-11'\n"
                "Error: '2019-05-0X' is not a valid date value, use ISO 8601 date\n"
                "Error: '3a' is not a valid months value, use a positive integer\n"
                "Error: invalid duration option 'x', allowed options are: "
                "'hours', 'monthdays', 'months', 'moon', 'weekdays', "
                "'weeks', 'weekyears', 'years', 'yearsdays'\n"
                "Error: Duplicate options in a single (sub)expression: 'months'\n"
                "Error: '7-X' is not a valid months value, use 1..12 or 1..11-2..12\n"
                "Error: '7-6' is not a valid weekdays value, use 1..7 or 1..6-2..7\n"
                "Error: '202a' is not a valid years value, use an integer or "
                "integer-integer\n"
                "Error: invalid datespec option 'x', allowed options are: "
                "'hours', 'monthdays', 'months', 'moon', 'weekdays', "
                "'weeks', 'weekyears', 'years', 'yearsdays'\n"
                "Error: Duplicate options in a single (sub)expression: 'years'\n"
                "Error: Errors have occurred, therefore pcs is unable to continue\n"
            ),
        )


class RscDefaultsSetCreate(
    get_assert_pcs_effect_mixin(
        lambda cib: etree.tostring(
            # pylint:disable=undefined-variable
            etree.parse(cib).findall(".//rsc_defaults")[0]
        )
    ),
    DefaultsSetCreateMixin,
    TestCase,
):
    cli_command = ["resource", "defaults"]
    cib_tag = "rsc_defaults"

    @skip_unless_pacemaker_supports_rsc_and_op_rules()
    def test_success_rules_rsc_op(self):
        self.assert_effect(
            self.cli_command
            + "set create id=X meta nam1=val1 rule resource ::Dummy".split(),
            f"""\
            <{self.cib_tag}>
                <meta_attributes id="X">
                    <rule id="X-rule" boolean-op="and" score="INFINITY">
                        <rsc_expression id="X-rule-rsc-Dummy" type="Dummy"/>
                    </rule>
                    <nvpair id="X-nam1" name="nam1" value="val1"/>
                </meta_attributes>
            </{self.cib_tag}>
            """,
            output=(
                "CIB has been upgraded to the latest schema version.\n"
                "Warning: Defaults do not apply to resources which override "
                "them with their own defined values\n"
            ),
        )

    def test_node_attr_expressions(self):
        self.assert_pcs_fail(
            self.cli_command + ("set create rule defined attr").split(),
            (
                "Error: Keywords 'defined', 'not_defined', 'eq', 'ne', 'gte', "
                "'gt', 'lte' and 'lt' cannot be used in a rule in this command\n"
                "Error: Errors have occurred, therefore pcs is unable to continue\n"
            ),
        )


class OpDefaultsSetCreate(
    get_assert_pcs_effect_mixin(
        lambda cib: etree.tostring(
            # pylint:disable=undefined-variable
            etree.parse(cib).findall(".//op_defaults")[0]
        )
    ),
    DefaultsSetCreateMixin,
    TestCase,
):
    cli_command = ["resource", "op", "defaults"]
    cib_tag = "op_defaults"

    def test_rule_error_messages(self):
        self.assert_pcs_fail(
            self.cli_command
            + (
                "set create rule defined attr1 or attr2 gte number 12a or "
                "attr3 lt version 3.2.1a or attr4 ne string test or attr5 lt 3 "
            ).split(),
            (
                "Error: '12a' is not a valid number attribute value, use a "
                "floating-point number\n"
                "Error: '3.2.1a' is not a valid version attribute value, use "
                "a version number (e.g. 1, 1.2, 1.23.45, ...)\n"
                "Error: Errors have occurred, therefore pcs is unable to continue\n"
            ),
        )

    @skip_unless_pacemaker_supports_rsc_and_op_rules()
    def test_success_rules_rsc_op(self):
        self.assert_effect(
            self.cli_command
            + (
                "-- set create id=X meta nam1=val1 "
                "rule resource ::Dummy and (op start or op stop) and "
                "(defined attr1 or attr2 gte number -1.2 or "
                "attr3 lt version 3.2.1 or attr4 ne string test or attr5 lt 3) "
            ).split(),
            f"""\
            <{self.cib_tag}>
                <meta_attributes id="X">
                    <rule id="X-rule" boolean-op="and" score="INFINITY">
                        <rsc_expression id="X-rule-rsc-Dummy" type="Dummy"/>
                        <rule id="X-rule-rule" boolean-op="or" score="0">
                            <op_expression id="X-rule-rule-op-start"
                                name="start"
                            />
                            <op_expression id="X-rule-rule-op-stop"
                                name="stop"
                            />
                        </rule>
                        <rule id="X-rule-rule-1" boolean-op="or" score="0">
                            <expression id="X-rule-rule-1-expr"
                                operation="defined" attribute="attr1"
                            />
                            <expression id="X-rule-rule-1-expr-1"
                                attribute="attr2" operation="gte"
                                type="number" value="-1.2"
                            />
                            <expression id="X-rule-rule-1-expr-2"
                                attribute="attr3" operation="lt"
                                type="version" value="3.2.1"
                            />
                            <expression id="X-rule-rule-1-expr-3"
                                attribute="attr4" operation="ne"
                                type="string" value="test"
                            />
                            <expression id="X-rule-rule-1-expr-4"
                                attribute="attr5" operation="lt" value="3"
                            />
                        </rule>
                    </rule>
                    <nvpair id="X-nam1" name="nam1" value="val1"/>
                </meta_attributes>
            </{self.cib_tag}>
            """,
            output=(
                "CIB has been upgraded to the latest schema version.\n"
                "Warning: Defaults do not apply to resources which override "
                "them with their own defined values\n"
            ),
        )


class DefaultsSetDeleteMixin(TestDefaultsMixin, AssertPcsMixin):
    cli_command = []
    prefix = ""
    cib_tag = ""

    def setUp(self):
        super().setUp()
        xml_rsc = """
            <rsc_defaults>
                <meta_attributes id="rsc-set1" />
                <meta_attributes id="rsc-set2" />
                <meta_attributes id="rsc-set3" />
                <meta_attributes id="rsc-set4" />
            </rsc_defaults>
        """
        xml_op = """
            <op_defaults>
                <meta_attributes id="op-set1" />
                <meta_attributes id="op-set2" />
                <meta_attributes id="op-set3" />
                <meta_attributes id="op-set4" />
            </op_defaults>
        """
        xml_manip = XmlManipulation.from_file(empty_cib)
        xml_manip.append_to_first_tag_name("configuration", xml_rsc, xml_op)
        write_data_to_tmpfile(str(xml_manip), self.temp_cib)

    def test_success(self):
        self.assert_effect(
            [
                self.cli_command
                + f"set delete {self.prefix}-set1 {self.prefix}-set3".split(),
                self.cli_command
                + f"set remove {self.prefix}-set1 {self.prefix}-set3".split(),
            ],
            dedent(
                f"""\
                <{self.cib_tag}>
                    <meta_attributes id="{self.prefix}-set2" />
                    <meta_attributes id="{self.prefix}-set4" />
                </{self.cib_tag}>
            """
            ),
        )


class RscDefaultsSetDelete(
    get_assert_pcs_effect_mixin(
        lambda cib: etree.tostring(
            # pylint:disable=undefined-variable
            etree.parse(cib).findall(".//rsc_defaults")[0]
        )
    ),
    DefaultsSetDeleteMixin,
    TestCase,
):
    cli_command = ["resource", "defaults"]
    prefix = "rsc"
    cib_tag = "rsc_defaults"


class OpDefaultsSetDelete(
    get_assert_pcs_effect_mixin(
        lambda cib: etree.tostring(
            # pylint:disable=undefined-variable
            etree.parse(cib).findall(".//op_defaults")[0]
        )
    ),
    DefaultsSetDeleteMixin,
    TestCase,
):
    cli_command = ["resource", "op", "defaults"]
    prefix = "op"
    cib_tag = "op_defaults"


class DefaultsSetUpdateMixin(TestDefaultsMixin, AssertPcsMixin):
    cli_command = []
    prefix = ""
    cib_tag = ""

    def test_success(self):
        xml = f"""
            <{self.cib_tag}>
                <meta_attributes id="my-set">
                    <nvpair id="my-set-name1" name="name1" value="value1" />
                    <nvpair id="my-set-name2" name="name2" value="value2" />
                    <nvpair id="my-set-name3" name="name3" value="value3" />
                </meta_attributes>
            </{self.cib_tag}>
        """
        xml_manip = XmlManipulation.from_file(empty_cib)
        xml_manip.append_to_first_tag_name("configuration", xml)
        write_data_to_tmpfile(str(xml_manip), self.temp_cib)
        warnings = (
            "Warning: Defaults do not apply to resources which override "
            "them with their own defined values\n"
        )

        self.assert_effect(
            self.cli_command
            + "set update my-set meta name2=value2A name3=".split(),
            dedent(
                f"""\
                <{self.cib_tag}>
                    <meta_attributes id="my-set">
                        <nvpair id="my-set-name1" name="name1" value="value1" />
                        <nvpair id="my-set-name2" name="name2" value="value2A" />
                    </meta_attributes>
                </{self.cib_tag}>
            """
            ),
            output=warnings,
        )

        self.assert_effect(
            self.cli_command + "set update my-set meta name1= name2=".split(),
            dedent(
                f"""\
                <{self.cib_tag}>
                    <meta_attributes id="my-set" />
                </{self.cib_tag}>
            """
            ),
            output=warnings,
        )


class RscDefaultsSetUpdate(
    get_assert_pcs_effect_mixin(
        lambda cib: etree.tostring(
            # pylint:disable=undefined-variable
            etree.parse(cib).findall(".//rsc_defaults")[0]
        )
    ),
    DefaultsSetUpdateMixin,
    TestCase,
):
    cli_command = ["resource", "defaults"]
    prefix = "rsc"
    cib_tag = "rsc_defaults"


class OpDefaultsSetUpdate(
    get_assert_pcs_effect_mixin(
        lambda cib: etree.tostring(
            # pylint:disable=undefined-variable
            etree.parse(cib).findall(".//op_defaults")[0]
        )
    ),
    DefaultsSetUpdateMixin,
    TestCase,
):
    cli_command = ["resource", "op", "defaults"]
    prefix = "op"
    cib_tag = "op_defaults"


class DefaultsSetUsageMixin(TestDefaultsMixin, AssertPcsMixin):
    cli_command = []

    def test_no_args(self):
        self.assert_pcs_fail(
            self.cli_command + ["set"],
            stdout_start=f"\nUsage: pcs {' '.join(self.cli_command)} set...\n",
        )

    def test_bad_command(self):
        self.assert_pcs_fail(
            self.cli_command + ["set", "bad-command"],
            stdout_start=f"\nUsage: pcs {' '.join(self.cli_command)} set ...\n",
        )


class RscDefaultsSetUsage(
    DefaultsSetUsageMixin,
    TestCase,
):
    cli_command = ["resource", "defaults"]


class OpDefaultsSetUsage(
    DefaultsSetUsageMixin,
    TestCase,
):
    cli_command = ["resource", "op", "defaults"]


class DefaultsUpdateMixin(TestDefaultsMixin, AssertPcsMixin):
    cli_command = []
    prefix = ""
    cib_tag = ""

    def assert_success_legacy(self, update_keyword):
        write_file_to_tmpfile(empty_cib, self.temp_cib)
        warning_lines = []
        if not update_keyword:
            warning_lines.append(
                f"Deprecation Warning: This command is deprecated and will be "
                f"removed. Please use 'pcs {' '.join(self.cli_command)} "
                f"update' instead.\n"
            )
        warning_lines.append(
            "Warning: Defaults do not apply to resources which override "
            "them with their own defined values\n"
        )
        warnings = "".join(warning_lines)

        command = self.cli_command[:]
        if update_keyword:
            command.append("update")

        self.assert_effect(
            command + "name1=value1 name2=value2 name3=value3".split(),
            dedent(
                f"""\
                <{self.cib_tag}>
                    <meta_attributes id="{self.cib_tag}-meta_attributes">
                        <nvpair id="{self.cib_tag}-meta_attributes-name1"
                            name="name1" value="value1"
                        />
                        <nvpair id="{self.cib_tag}-meta_attributes-name2"
                            name="name2" value="value2"
                        />
                        <nvpair id="{self.cib_tag}-meta_attributes-name3"
                            name="name3" value="value3"
                        />
                    </meta_attributes>
                </{self.cib_tag}>
            """
            ),
            output=warnings,
        )

        self.assert_effect(
            command + "name2=value2A name3=".split(),
            dedent(
                f"""\
                <{self.cib_tag}>
                    <meta_attributes id="{self.cib_tag}-meta_attributes">
                        <nvpair id="{self.cib_tag}-meta_attributes-name1"
                            name="name1" value="value1"
                        />
                        <nvpair id="{self.cib_tag}-meta_attributes-name2"
                            name="name2" value="value2A"
                        />
                    </meta_attributes>
                </{self.cib_tag}>
            """
            ),
            output=warnings,
        )

        self.assert_effect(
            command + "name1= name2=".split(),
            dedent(
                f"""\
                <{self.cib_tag}>
                    <meta_attributes id="{self.cib_tag}-meta_attributes" />
                </{self.cib_tag}>
            """
            ),
            output=warnings,
        )

    def test_deprecated(self):
        self.assert_success_legacy(False)

    def test_legacy(self):
        self.assert_success_legacy(True)


class RscDefaultsUpdate(
    get_assert_pcs_effect_mixin(
        lambda cib: etree.tostring(
            # pylint:disable=undefined-variable
            etree.parse(cib).findall(".//rsc_defaults")[0]
        )
    ),
    DefaultsUpdateMixin,
    TestCase,
):
    cli_command = ["resource", "defaults"]
    prefix = "rsc"
    cib_tag = "rsc_defaults"


class OpDefaultsUpdate(
    get_assert_pcs_effect_mixin(
        lambda cib: etree.tostring(
            # pylint:disable=undefined-variable
            etree.parse(cib).findall(".//op_defaults")[0]
        )
    ),
    DefaultsUpdateMixin,
    TestCase,
):
    cli_command = ["resource", "op", "defaults"]
    prefix = "op"
    cib_tag = "op_defaults"



import os
import tempfile
import uuid
from contextlib import contextmanager
from typing import (
    IO,
    Callable,
    ContextManager,
    Generator,
    Literal,
    Mapping,
    Optional,
    TypeVar,
    Union,
    overload,
)

from pcs.common import reports
from pcs.lib.errors import LibraryError

T = TypeVar("T")


def get_optional_value(
    constructor: Callable[[str], T], value: Optional[str]
) -> Optional[T]:
    if value is None:
        return None
    return constructor(value)


def generate_binary_key(random_bytes_count: int) -> bytes:
    return os.urandom(random_bytes_count)


def generate_uuid() -> str:
    return uuid.uuid4().hex


def environment_file_to_dict(config: str) -> dict[str, str]:
    """
    Parse systemd Environment file. This parser is simplified version of
    parser in systemd, because of their poor implementation.
    Returns configuration in dictionary in format:
    {
        <option>: <value>,
        ...
    }

    config -- Environment file as string
    """
    # escape new lines
    config = config.replace("\\\n", "")

    data = {}
    for line in [l.strip() for l in config.split("\n")]:
        if line == "" or line.startswith("#") or line.startswith(";"):
            continue
        if "=" not in line:
            continue
        key, val = line.split("=", 1)
        value = val.strip()
        data[key.strip()] = value
    return data


def dict_to_environment_file(config_dict: Mapping[str, str]) -> str:
    """
    Convert data in dictionary to Environment file format.
    Returns Environment file as string in format:
    # comment
    <option>=<value>
    ...

    config_dict -- dictionary in format: { <option>: <value>, ...}
    """
    lines = ["# This file has been generated by pcs.\n"]
    for key, val in sorted(config_dict.items()):
        lines.append("{key}={val}\n".format(key=key, val=val))
    return "".join(lines)


@overload
def get_tmp_file(
    data: Optional[bytes], binary: Literal[True]
) -> ContextManager[IO[bytes]]:
    pass


@overload
def get_tmp_file(
    data: Optional[str],
    binary: Literal[False] = False,
) -> ContextManager[IO[str]]:
    pass


# We ignore return type here as it doesn't work with mypy (@contextmanager and
# @overload) and it is properly typed in @overload functions.
@contextmanager
def get_tmp_file(  # type: ignore
    data: Optional[Union[bytes, str]],
    binary: bool = False,
):
    mode = "w+b" if binary else "w+"
    tmpfile = None
    try:
        with tempfile.NamedTemporaryFile(mode=mode, suffix=".pcs") as tmpfile:
            if data is not None:
                tmpfile.write(data)
                tmpfile.flush()
            yield tmpfile
    finally:
        if tmpfile:
            tmpfile.close()


@contextmanager
def get_tmp_cib(
    report_processor: reports.ReportProcessor, data: Optional[str]
) -> Generator[IO[str], None, None]:
    try:
        with get_tmp_file(data) as tmp_cib_file:
            report_processor.report(
                reports.ReportItem.debug(
                    reports.messages.TmpFileWrite(tmp_cib_file.name, data or "")
                )
            )
            yield tmp_cib_file
    except EnvironmentError as e:
        raise LibraryError(
            reports.ReportItem.error(reports.messages.CibSaveTmpError(str(e)))
        ) from e


def create_tmp_cib(
    report_processor: reports.ReportProcessor, data: Optional[str]
) -> IO[str]:
    try:
        # pylint: disable=consider-using-with
        tmp_file = tempfile.NamedTemporaryFile(mode="w+", suffix=".pcs")
        if data is not None:
            tmp_file.write(data)
            tmp_file.flush()
        report_processor.report(
            reports.ReportItem.debug(
                reports.messages.TmpFileWrite(tmp_file.name, data or "")
            )
        )
        return tmp_file
    except EnvironmentError as e:
        raise LibraryError(
            reports.ReportItem.error(reports.messages.CibSaveTmpError(str(e)))
        ) from e



import re
from textwrap import dedent
from unittest import TestCase

from pcs.cli import rule
from pcs.common.pacemaker.rule import (
    CibRuleDateCommonDto,
    CibRuleExpressionDto,
)
from pcs.common.types import (
    CibRuleExpressionType,
    CibRuleInEffectStatus,
)


class RuleDtoToLinesMixin:
    @staticmethod
    def _export(dto, with_ids):
        return (
            "\n".join(rule.rule_expression_dto_to_lines(dto, with_ids=with_ids))
            + "\n"
        )

    def assert_lines(self, dto, lines):
        self.assertEqual(
            self._export(dto, True),
            lines,
        )
        self.assertEqual(
            self._export(dto, False),
            re.sub(r" +\(id:.*\)", "", lines),
        )


class ExpressionDtoToLines(RuleDtoToLinesMixin, TestCase):
    def test_defined(self):
        dto = CibRuleExpressionDto(
            "my-id",
            CibRuleExpressionType.RULE,
            CibRuleInEffectStatus.UNKNOWN,
            {},
            None,
            None,
            [
                CibRuleExpressionDto(
                    "my-id-expr",
                    CibRuleExpressionType.EXPRESSION,
                    CibRuleInEffectStatus.UNKNOWN,
                    {"attribute": "pingd", "operation": "defined"},
                    None,
                    None,
                    [],
                    "defined pingd",
                ),
            ],
            "defined pingd",
        )
        output = dedent(
            """\
              Rule: (id:my-id)
                Expression: defined pingd (id:my-id-expr)
            """
        )
        self.assert_lines(dto, output)

    def test_value_comparison(self):
        dto = CibRuleExpressionDto(
            "my-id",
            CibRuleExpressionType.RULE,
            CibRuleInEffectStatus.UNKNOWN,
            {},
            None,
            None,
            [
                CibRuleExpressionDto(
                    "my-id-expr",
                    CibRuleExpressionType.EXPRESSION,
                    CibRuleInEffectStatus.UNKNOWN,
                    {
                        "attribute": "my-attr",
                        "operation": "eq",
                        "value": "my value",
                    },
                    None,
                    None,
                    [],
                    "my-attr eq 'my value'",
                ),
            ],
            "my-attr eq 'my value'",
        )
        output = dedent(
            """\
              Rule: (id:my-id)
                Expression: my-attr eq 'my value' (id:my-id-expr)
            """
        )
        self.assert_lines(dto, output)

    def test_value_comparison_with_type(self):
        dto = CibRuleExpressionDto(
            "my-id",
            CibRuleExpressionType.RULE,
            CibRuleInEffectStatus.UNKNOWN,
            {},
            None,
            None,
            [
                CibRuleExpressionDto(
                    "my-id-expr",
                    CibRuleExpressionType.EXPRESSION,
                    CibRuleInEffectStatus.UNKNOWN,
                    {
                        "attribute": "foo",
                        "operation": "gt",
                        "type": "version",
                        "value": "1.2.3",
                    },
                    None,
                    None,
                    [],
                    "foo gt version 1.2.3",
                ),
            ],
            "foo gt version 1.2.3",
        )
        output = dedent(
            """\
              Rule: (id:my-id)
                Expression: foo gt version 1.2.3 (id:my-id-expr)
            """
        )
        self.assert_lines(dto, output)


class DateExpressionDtoToLines(RuleDtoToLinesMixin, TestCase):
    def test_simple(self):
        dto = CibRuleExpressionDto(
            "rule",
            CibRuleExpressionType.RULE,
            CibRuleInEffectStatus.UNKNOWN,
            {},
            None,
            None,
            [
                CibRuleExpressionDto(
                    "rule-expr",
                    CibRuleExpressionType.DATE_EXPRESSION,
                    CibRuleInEffectStatus.UNKNOWN,
                    {"operation": "gt", "start": "2014-06-26"},
                    None,
                    None,
                    [],
                    "date gt 2014-06-26",
                ),
            ],
            "date gt 2014-06-26",
        )
        output = dedent(
            """\
              Rule: (id:rule)
                Expression: date gt 2014-06-26 (id:rule-expr)
            """
        )
        self.assert_lines(dto, output)

    def test_datespec(self):
        dto = CibRuleExpressionDto(
            "rule",
            CibRuleExpressionType.RULE,
            CibRuleInEffectStatus.UNKNOWN,
            {},
            None,
            None,
            [
                CibRuleExpressionDto(
                    "rule-expr",
                    CibRuleExpressionType.DATE_EXPRESSION,
                    CibRuleInEffectStatus.UNKNOWN,
                    {"operation": "date_spec"},
                    CibRuleDateCommonDto(
                        "rule-expr-datespec",
                        {"hours": "1-14", "monthdays": "20-30", "months": "1"},
                    ),
                    None,
                    [],
                    "date-spec hours=1-14 monthdays=20-30 months=1",
                ),
            ],
            "date-spec hours=1-14 monthdays=20-30 months=1",
        )
        output = dedent(
            """\
              Rule: (id:rule)
                Expression: (id:rule-expr)
                  Date Spec: hours=1-14 monthdays=20-30 months=1 (id:rule-expr-datespec)
            """
        )
        self.assert_lines(dto, output)

    def test_inrange_start_end(self):
        dto = CibRuleExpressionDto(
            "rule",
            CibRuleExpressionType.RULE,
            CibRuleInEffectStatus.UNKNOWN,
            {},
            None,
            None,
            [
                CibRuleExpressionDto(
                    "rule-expr",
                    CibRuleExpressionType.DATE_EXPRESSION,
                    CibRuleInEffectStatus.UNKNOWN,
                    {
                        "operation": "in_range",
                        "start": "2014-06-26",
                        "end": "2014-07-26",
                    },
                    None,
                    None,
                    [],
                    "date in_range 2014-06-26 to 2014-07-26",
                ),
            ],
            "date in_range 2014-06-26 to 2014-07-26",
        )
        output = dedent(
            """\
              Rule: (id:rule)
                Expression: date in_range 2014-06-26 to 2014-07-26 (id:rule-expr)
            """
        )
        self.assert_lines(dto, output)

    def test_inrange_end(self):
        dto = CibRuleExpressionDto(
            "rule",
            CibRuleExpressionType.RULE,
            CibRuleInEffectStatus.UNKNOWN,
            {},
            None,
            None,
            [
                CibRuleExpressionDto(
                    "rule-expr",
                    CibRuleExpressionType.DATE_EXPRESSION,
                    CibRuleInEffectStatus.UNKNOWN,
                    {"operation": "in_range", "end": "2014-07-26"},
                    None,
                    None,
                    [],
                    "date in_range to 2014-07-26",
                ),
            ],
            "date in_range to 2014-07-26",
        )
        output = dedent(
            """\
              Rule: (id:rule)
                Expression: date in_range to 2014-07-26 (id:rule-expr)
            """
        )
        self.assert_lines(dto, output)

    def test_inrange_start_duration(self):
        dto = CibRuleExpressionDto(
            "rule",
            CibRuleExpressionType.RULE,
            CibRuleInEffectStatus.UNKNOWN,
            {},
            None,
            None,
            [
                CibRuleExpressionDto(
                    "rule-expr",
                    CibRuleExpressionType.DATE_EXPRESSION,
                    CibRuleInEffectStatus.UNKNOWN,
                    {
                        "operation": "in_range",
                        "start": "2014-06-26",
                    },
                    None,
                    CibRuleDateCommonDto("rule-expr-duration", {"years": "1"}),
                    [],
                    "date in_range 2014-06-26 to duration years=1",
                ),
            ],
            "date in_range 2014-06-26 to duration years=1",
        )
        output = dedent(
            """\
              Rule: (id:rule)
                Expression: date in_range 2014-06-26 to duration (id:rule-expr)
                  Duration: years=1 (id:rule-expr-duration)
            """
        )
        self.assert_lines(dto, output)


class OpExpressionDtoToLines(RuleDtoToLinesMixin, TestCase):
    def test_minimal(self):
        dto = CibRuleExpressionDto(
            "my-id",
            CibRuleExpressionType.RULE,
            CibRuleInEffectStatus.UNKNOWN,
            {},
            None,
            None,
            [
                CibRuleExpressionDto(
                    "my-id-op",
                    CibRuleExpressionType.OP_EXPRESSION,
                    CibRuleInEffectStatus.UNKNOWN,
                    {"name": "start"},
                    None,
                    None,
                    [],
                    "op start",
                ),
            ],
            "op start",
        )
        output = dedent(
            """\
              Rule: (id:my-id)
                Expression: op start (id:my-id-op)
            """
        )
        self.assert_lines(dto, output)

    def test_interval(self):
        dto = CibRuleExpressionDto(
            "my-id",
            CibRuleExpressionType.RULE,
            CibRuleInEffectStatus.UNKNOWN,
            {},
            None,
            None,
            [
                CibRuleExpressionDto(
                    "my-id-op",
                    CibRuleExpressionType.OP_EXPRESSION,
                    CibRuleInEffectStatus.UNKNOWN,
                    {"name": "start", "interval": "2min"},
                    None,
                    None,
                    [],
                    "op start interval=2min",
                ),
            ],
            "op start interval=2min",
        )
        output = dedent(
            """\
              Rule: (id:my-id)
                Expression: op start interval=2min (id:my-id-op)
            """
        )
        self.assert_lines(dto, output)


class ResourceExpressionDtoToLines(RuleDtoToLinesMixin, TestCase):
    def test_success(self):
        dto = CibRuleExpressionDto(
            "my-id",
            CibRuleExpressionType.RULE,
            CibRuleInEffectStatus.UNKNOWN,
            {},
            None,
            None,
            [
                CibRuleExpressionDto(
                    "my-id-expr",
                    CibRuleExpressionType.RSC_EXPRESSION,
                    CibRuleInEffectStatus.UNKNOWN,
                    {"class": "ocf", "provider": "pacemaker", "type": "Dummy"},
                    None,
                    None,
                    [],
                    "resource ocf:pacemaker:Dummy",
                ),
            ],
            "resource ocf:pacemaker:Dummy",
        )
        output = dedent(
            """\
              Rule: (id:my-id)
                Expression: resource ocf:pacemaker:Dummy (id:my-id-expr)
            """
        )
        self.assert_lines(dto, output)


class InEffect(RuleDtoToLinesMixin, TestCase):
    @staticmethod
    def fixture_dto(expired):
        return CibRuleExpressionDto(
            "my-id",
            CibRuleExpressionType.RULE,
            expired,
            {},
            None,
            None,
            [
                CibRuleExpressionDto(
                    "my-id-expr",
                    CibRuleExpressionType.EXPRESSION,
                    CibRuleInEffectStatus.UNKNOWN,
                    {"attribute": "pingd", "operation": "defined"},
                    None,
                    None,
                    [],
                    "defined pingd",
                ),
            ],
            "defined pingd",
        )

    def test_unknown(self):
        self.assert_lines(
            self.fixture_dto(CibRuleInEffectStatus.UNKNOWN),
            dedent(
                """\
                  Rule: (id:my-id)
                    Expression: defined pingd (id:my-id-expr)
                """
            ),
        )

    def test_expired(self):
        self.assert_lines(
            self.fixture_dto(CibRuleInEffectStatus.EXPIRED),
            dedent(
                """\
                  Rule (expired): (id:my-id)
                    Expression: defined pingd (id:my-id-expr)
                """
            ),
        )

    def test_not_yet_effective(self):
        self.assert_lines(
            self.fixture_dto(CibRuleInEffectStatus.NOT_YET_IN_EFFECT),
            dedent(
                """\
                  Rule (not yet in effect): (id:my-id)
                    Expression: defined pingd (id:my-id-expr)
                """
            ),
        )


class RuleDtoToLines(RuleDtoToLinesMixin, TestCase):
    def test_complex_rule(self):
        dto = CibRuleExpressionDto(
            "complex",
            CibRuleExpressionType.RULE,
            CibRuleInEffectStatus.UNKNOWN,
            {"boolean-op": "or", "score": "INFINITY"},
            None,
            None,
            [
                CibRuleExpressionDto(
                    "complex-rule-1",
                    CibRuleExpressionType.RULE,
                    CibRuleInEffectStatus.UNKNOWN,
                    {"boolean-op": "and", "score": "0"},
                    None,
                    None,
                    [
                        CibRuleExpressionDto(
                            "complex-rule-1-expr",
                            CibRuleExpressionType.DATE_EXPRESSION,
                            CibRuleInEffectStatus.UNKNOWN,
                            {"operation": "date_spec"},
                            CibRuleDateCommonDto(
                                "complex-rule-1-expr-datespec",
                                {"hours": "12-23", "weekdays": "1-5"},
                            ),
                            None,
                            [],
                            "date-spec hours=12-23 weekdays=1-5",
                        ),
                        CibRuleExpressionDto(
                            "complex-rule-1-expr-1",
                            CibRuleExpressionType.DATE_EXPRESSION,
                            CibRuleInEffectStatus.UNKNOWN,
                            {
                                "operation": "in_range",
                                "start": "2014-07-26",
                            },
                            None,
                            CibRuleDateCommonDto(
                                "complex-rule-1-expr-1-durat",
                                {"months": "1"},
                            ),
                            [],
                            "date in_range 2014-07-26 to duration months=1",
                        ),
                    ],
                    "date-spec hours=12-23 weekdays=1-5 and date in_range "
                    "2014-07-26 to duration months=1",
                ),
                CibRuleExpressionDto(
                    "complex-rule",
                    CibRuleExpressionType.RULE,
                    CibRuleInEffectStatus.UNKNOWN,
                    {"boolean-op": "and", "score": "0"},
                    None,
                    None,
                    [
                        CibRuleExpressionDto(
                            "complex-rule-expr-1",
                            CibRuleExpressionType.EXPRESSION,
                            CibRuleInEffectStatus.UNKNOWN,
                            {
                                "attribute": "foo",
                                "operation": "gt",
                                "type": "version",
                                "value": "1.2",
                            },
                            None,
                            None,
                            [],
                            "foo gt version 1.2",
                        ),
                        CibRuleExpressionDto(
                            "complex-rule-expr",
                            CibRuleExpressionType.EXPRESSION,
                            CibRuleInEffectStatus.UNKNOWN,
                            {
                                "attribute": "#uname",
                                "operation": "eq",
                                "value": "node3 4",
                            },
                            None,
                            None,
                            [],
                            "#uname eq 'node3 4'",
                        ),
                        CibRuleExpressionDto(
                            "complex-rule-expr-2",
                            CibRuleExpressionType.EXPRESSION,
                            CibRuleInEffectStatus.UNKNOWN,
                            {
                                "attribute": "#uname",
                                "operation": "eq",
                                "value": "nodeA",
                            },
                            None,
                            None,
                            [],
                            "#uname eq nodeA",
                        ),
                    ],
                    "foo gt version 1.2 and #uname eq 'node3 4' and #uname "
                    "eq nodeA",
                ),
            ],
            "(date-spec hours=12-23 weekdays=1-5 and date in_range "
            "2014-07-26 to duration months=1) or (foo gt version 1.2 and "
            "#uname eq 'node3 4' and #uname eq nodeA)",
        )
        output = dedent(
            """\
            Rule: boolean-op=or score=INFINITY (id:complex)
              Rule: boolean-op=and score=0 (id:complex-rule-1)
                Expression: (id:complex-rule-1-expr)
                  Date Spec: hours=12-23 weekdays=1-5 (id:complex-rule-1-expr-datespec)
                Expression: date in_range 2014-07-26 to duration (id:complex-rule-1-expr-1)
                  Duration: months=1 (id:complex-rule-1-expr-1-durat)
              Rule: boolean-op=and score=0 (id:complex-rule)
                Expression: foo gt version 1.2 (id:complex-rule-expr-1)
                Expression: #uname eq 'node3 4' (id:complex-rule-expr)
                Expression: #uname eq nodeA (id:complex-rule-expr-2)
            """
        )
        self.assert_lines(dto, output)



import errno
import fcntl
import os.path
from unittest import (
    TestCase,
    mock,
)

from pcs.common.file import (
    FileAlreadyExists,
    FileMetadata,
    RawFile,
    RawFileError,
)

from pcs_test.tools.misc import (
    create_patcher,
    get_tmp_dir,
    get_tmp_file,
    write_data_to_tmpfile,
)

FILE_TYPE_CODE = "file type code"
FILE_PATH = "file path"
FILE_OWNER = "owner name"
FILE_GROUP = "group name"
FILE_PERMISSIONS = "file permissions"
FILE_IS_BINARY = False

patch_file = create_patcher("pcs.common.file")


def fixture_metadata(
    owner=FILE_OWNER,
    group=FILE_GROUP,
    permissions=FILE_PERMISSIONS,
    binary=FILE_IS_BINARY,
    file_path=FILE_PATH,
):
    return FileMetadata(
        FILE_TYPE_CODE, file_path, owner, group, permissions, binary
    )


class RawFileExists(TestCase):
    @patch_file("os.path.exists")
    def test_return_true_if_file_exists(self, mock_exists):
        mock_exists.return_value = True
        self.assertTrue(RawFile(fixture_metadata()).exists())
        mock_exists.assert_called_once_with(FILE_PATH)

    @patch_file("os.path.exists")
    def test_return_false_if_file_does_not_exist(self, mock_exists):
        mock_exists.return_value = False
        self.assertFalse(RawFile(fixture_metadata()).exists())
        mock_exists.assert_called_once_with(FILE_PATH)


@patch_file("fcntl.flock")
class RawFileRead(TestCase):
    def assert_read_in_correct_mode(self, mock_flock, raw_file, mode):
        file_data = "some file data"
        returned_data = file_data.encode("utf-8")
        if "b" in mode:
            file_data = file_data.encode("utf-8")
        mock_open = mock.mock_open()
        with patch_file("open", mock_open):
            mock_open().read.return_value = file_data
            mock_open().fileno.return_value = 123
            self.assertEqual(returned_data, raw_file.read())
            mock_open().read.assert_called_once_with()
        mock_open.assert_has_calls([mock.call(FILE_PATH, mode)])
        mock_flock.assert_called_once_with(123, fcntl.LOCK_SH)

    def test_success_text(self, mock_flock):
        self.assert_read_in_correct_mode(
            mock_flock,
            RawFile(fixture_metadata()),
            "r",
        )

    def test_success_binary(self, mock_flock):
        self.assert_read_in_correct_mode(
            mock_flock,
            RawFile(fixture_metadata(binary=True)),
            "rb",
        )

    def test_cannot_open(self, mock_flock):
        raw_file = RawFile(fixture_metadata())
        mock_open = mock.mock_open()
        with patch_file("open", mock_open):
            mock_open.side_effect = OSError(1, "some error", FILE_PATH)
            with self.assertRaises(RawFileError) as cm:
                raw_file.read()
                mock_open().read.assert_not_called()
        mock_open.assert_has_calls([mock.call(FILE_PATH, "r")])
        mock_flock.assert_not_called()
        self.assertEqual(cm.exception.metadata, raw_file.metadata)
        self.assertEqual(cm.exception.action, RawFileError.ACTION_READ)
        self.assertEqual(cm.exception.reason, f"some error: '{FILE_PATH}'")

    def test_cannot_lock(self, mock_flock):
        raw_file = RawFile(fixture_metadata())
        mock_flock.side_effect = OSError(1, "some error", FILE_PATH)
        mock_open = mock.mock_open()
        with patch_file("open", mock_open):
            mock_open().fileno.return_value = 123
            with self.assertRaises(RawFileError) as cm:
                raw_file.read()
                mock_open().read.assert_not_called()
        mock_open.assert_has_calls([mock.call(FILE_PATH, "r")])
        mock_flock.assert_called_once_with(123, fcntl.LOCK_SH)
        self.assertEqual(cm.exception.metadata, raw_file.metadata)
        self.assertEqual(cm.exception.action, RawFileError.ACTION_READ)
        self.assertEqual(cm.exception.reason, f"some error: '{FILE_PATH}'")

    def test_cannot_read(self, mock_flock):
        raw_file = RawFile(fixture_metadata())
        mock_open = mock.mock_open()
        with patch_file("open", mock_open):
            mock_open().read.side_effect = OSError(1, "some error", FILE_PATH)
            mock_open().fileno.return_value = 123
            with self.assertRaises(RawFileError) as cm:
                raw_file.read()
            mock_open().read.assert_called_once_with()
        mock_open.assert_has_calls([mock.call(FILE_PATH, "r")])
        mock_flock.assert_called_once_with(123, fcntl.LOCK_SH)
        self.assertEqual(cm.exception.metadata, raw_file.metadata)
        self.assertEqual(cm.exception.action, RawFileError.ACTION_READ)
        self.assertEqual(cm.exception.reason, f"some error: '{FILE_PATH}'")


@patch_file("fcntl.flock")
@patch_file("shutil.chown")
@patch_file("os.chmod")
class RawFileWrite(TestCase):
    fileno = 123

    def assert_success(self, mock_flock, raw_file, mode, can_overwrite):
        file_data = "some file data".encode("utf-8")
        written_data = file_data
        if not "b" in mode:
            written_data = written_data.decode("utf-8")
        mock_open = mock.mock_open()
        with patch_file("open", mock_open):
            mock_open().fileno.return_value = self.fileno
            raw_file.write(file_data, can_overwrite=can_overwrite)
            mock_open().write.assert_called_once_with(written_data)
        mock_open.assert_has_calls([mock.call(FILE_PATH, mode)])
        mock_flock.assert_called_once_with(self.fileno, fcntl.LOCK_EX)

    def test_success_minimal(self, mock_chmod, mock_chown, mock_flock):
        self.assert_success(
            mock_flock,
            RawFile(fixture_metadata(owner=None, group=None, permissions=None)),
            "x",
            False,
        )
        mock_chown.assert_not_called()
        mock_chmod.assert_not_called()

    def test_success_minimal_binary(self, mock_chmod, mock_chown, mock_flock):
        self.assert_success(
            mock_flock,
            RawFile(
                fixture_metadata(
                    owner=None, group=None, permissions=None, binary=True
                )
            ),
            "xb",
            False,
        )
        mock_chown.assert_not_called()
        mock_chmod.assert_not_called()

    def test_success_overwrite(self, mock_chmod, mock_chown, mock_flock):
        self.assert_success(
            mock_flock,
            RawFile(fixture_metadata(owner=None, group=None, permissions=None)),
            "w",
            True,
        )
        mock_chown.assert_not_called()
        mock_chmod.assert_not_called()

    def test_success_overwrite_binary(self, mock_chmod, mock_chown, mock_flock):
        self.assert_success(
            mock_flock,
            RawFile(
                fixture_metadata(
                    owner=None, group=None, permissions=None, binary=True
                )
            ),
            "wb",
            True,
        )
        mock_chown.assert_not_called()
        mock_chmod.assert_not_called()

    def test_success_owner_user(self, mock_chmod, mock_chown, mock_flock):
        self.assert_success(
            mock_flock,
            RawFile(fixture_metadata(group=None, permissions=None)),
            "x",
            False,
        )
        mock_chown.assert_called_once_with(FILE_PATH, user=FILE_OWNER)
        mock_chmod.assert_not_called()

    def test_success_owner_group(self, mock_chmod, mock_chown, mock_flock):
        self.assert_success(
            mock_flock,
            RawFile(fixture_metadata(owner=None, permissions=None)),
            "x",
            False,
        )
        mock_chown.assert_called_once_with(FILE_PATH, group=FILE_GROUP)
        mock_chmod.assert_not_called()

    def test_success_owner_both(self, mock_chmod, mock_chown, mock_flock):
        self.assert_success(
            mock_flock, RawFile(fixture_metadata(permissions=None)), "x", False
        )
        mock_chown.assert_has_calls(
            [
                mock.call(FILE_PATH, user=FILE_OWNER),
                mock.call(FILE_PATH, group=FILE_GROUP),
            ]
        )
        self.assertEqual(mock_chown.call_count, 2)
        mock_chmod.assert_not_called()

    def test_success_permissions(self, mock_chmod, mock_chown, mock_flock):
        self.assert_success(
            mock_flock,
            RawFile(fixture_metadata(owner=None, group=None)),
            "x",
            False,
        )
        mock_chown.assert_not_called()
        mock_chmod.assert_called_once_with(self.fileno, FILE_PERMISSIONS)

    def test_success_chown_permissions(
        self, mock_chmod, mock_chown, mock_flock
    ):
        self.assert_success(mock_flock, RawFile(fixture_metadata()), "x", False)
        mock_chown.assert_has_calls(
            [
                mock.call(FILE_PATH, user=FILE_OWNER),
                mock.call(FILE_PATH, group=FILE_GROUP),
            ]
        )
        self.assertEqual(mock_chown.call_count, 2)
        mock_chmod.assert_called_once_with(self.fileno, FILE_PERMISSIONS)

    def test_already_exists(self, mock_chmod, mock_chown, mock_flock):
        raw_file = RawFile(fixture_metadata())
        mock_open = mock.mock_open()
        with patch_file("open", mock_open):
            mock_open.side_effect = OSError(
                errno.EEXIST, "some error", FILE_PATH
            )
            with self.assertRaises(FileAlreadyExists) as cm:
                raw_file.write(b"data")
                mock_open().write.assert_not_called()
        mock_open.assert_has_calls([mock.call(FILE_PATH, "x")])
        mock_flock.assert_not_called()
        mock_chown.assert_not_called()
        mock_chmod.assert_not_called()
        self.assertEqual(cm.exception.metadata, raw_file.metadata)
        self.assertEqual(cm.exception.action, RawFileError.ACTION_WRITE)
        self.assertEqual(cm.exception.reason, "")

    def test_cannot_open(self, mock_chmod, mock_chown, mock_flock):
        raw_file = RawFile(fixture_metadata())
        mock_open = mock.mock_open()
        with patch_file("open", mock_open):
            mock_open.side_effect = OSError(1, "some error", FILE_PATH)
            with self.assertRaises(RawFileError) as cm:
                raw_file.write(b"data")
                mock_open().write.assert_not_called()
        mock_open.assert_has_calls([mock.call(FILE_PATH, "x")])
        mock_flock.assert_not_called()
        mock_chown.assert_not_called()
        mock_chmod.assert_not_called()
        self.assertEqual(cm.exception.metadata, raw_file.metadata)
        self.assertEqual(cm.exception.action, RawFileError.ACTION_WRITE)
        self.assertEqual(cm.exception.reason, f"some error: '{FILE_PATH}'")

    def test_cannot_lock(self, mock_chmod, mock_chown, mock_flock):
        raw_file = RawFile(fixture_metadata())
        mock_flock.side_effect = OSError(1, "some error", FILE_PATH)
        mock_open = mock.mock_open()
        with patch_file("open", mock_open):
            mock_open().fileno.return_value = self.fileno
            with self.assertRaises(RawFileError) as cm:
                raw_file.write(b"data")
                mock_open().write.assert_not_called()
        mock_open.assert_has_calls([mock.call(FILE_PATH, "x")])
        mock_flock.assert_called_once_with(self.fileno, fcntl.LOCK_EX)
        mock_chown.assert_not_called()
        mock_chmod.assert_not_called()
        self.assertEqual(cm.exception.metadata, raw_file.metadata)
        self.assertEqual(cm.exception.action, RawFileError.ACTION_WRITE)
        self.assertEqual(cm.exception.reason, f"some error: '{FILE_PATH}'")

    def test_cannot_chown_bad_user(self, mock_chmod, mock_chown, mock_flock):
        raw_file = RawFile(fixture_metadata())
        mock_chown.side_effect = LookupError("some error")
        mock_open = mock.mock_open()
        with patch_file("open", mock_open):
            mock_open().fileno.return_value = self.fileno
            with self.assertRaises(RawFileError) as cm:
                raw_file.write(b"data")
                mock_open().write.assert_not_called()
        mock_open.assert_has_calls([mock.call(FILE_PATH, "x")])
        mock_flock.assert_called_once_with(self.fileno, fcntl.LOCK_EX)
        mock_chown.assert_called_once_with(FILE_PATH, user=FILE_OWNER)
        mock_chmod.assert_not_called()
        self.assertEqual(cm.exception.metadata, raw_file.metadata)
        self.assertEqual(cm.exception.action, RawFileError.ACTION_CHOWN)
        self.assertEqual(cm.exception.reason, "some error")

    def test_cannot_chown(self, mock_chmod, mock_chown, mock_flock):
        raw_file = RawFile(fixture_metadata())
        mock_chown.side_effect = OSError(1, "some error", FILE_PATH)
        mock_open = mock.mock_open()
        with patch_file("open", mock_open):
            mock_open().fileno.return_value = self.fileno
            with self.assertRaises(RawFileError) as cm:
                raw_file.write(b"data")
                mock_open().write.assert_not_called()
        mock_open.assert_has_calls([mock.call(FILE_PATH, "x")])
        mock_flock.assert_called_once_with(self.fileno, fcntl.LOCK_EX)
        mock_chown.assert_called_once_with(FILE_PATH, user=FILE_OWNER)
        mock_chmod.assert_not_called()
        self.assertEqual(cm.exception.metadata, raw_file.metadata)
        self.assertEqual(cm.exception.action, RawFileError.ACTION_CHOWN)
        self.assertEqual(cm.exception.reason, f"some error: '{FILE_PATH}'")

    def test_cannot_chmod(self, mock_chmod, mock_chown, mock_flock):
        raw_file = RawFile(fixture_metadata())
        mock_chmod.side_effect = OSError(1, "some error", FILE_PATH)
        mock_open = mock.mock_open()
        with patch_file("open", mock_open):
            mock_open().fileno.return_value = self.fileno
            with self.assertRaises(RawFileError) as cm:
                raw_file.write(b"data")
                mock_open().write.assert_not_called()
        mock_open.assert_has_calls([mock.call(FILE_PATH, "x")])
        mock_flock.assert_called_once_with(self.fileno, fcntl.LOCK_EX)
        mock_chown.assert_has_calls(
            [
                mock.call(FILE_PATH, user=FILE_OWNER),
                mock.call(FILE_PATH, group=FILE_GROUP),
            ]
        )
        self.assertEqual(mock_chown.call_count, 2)
        mock_chmod.assert_called_once_with(self.fileno, FILE_PERMISSIONS)
        self.assertEqual(cm.exception.metadata, raw_file.metadata)
        self.assertEqual(cm.exception.action, RawFileError.ACTION_CHMOD)
        self.assertEqual(cm.exception.reason, f"some error: '{FILE_PATH}'")

    def test_cannot_write(self, mock_chmod, mock_chown, mock_flock):
        raw_file = RawFile(fixture_metadata())
        mock_open = mock.mock_open()
        with patch_file("open", mock_open):
            mock_open().fileno.return_value = self.fileno
            mock_open().write.side_effect = OSError(1, "some error", FILE_PATH)
            with self.assertRaises(RawFileError) as cm:
                raw_file.write(b"data")
            mock_open().write.assert_called_once_with("data")
        mock_open.assert_has_calls([mock.call(FILE_PATH, "x")])
        mock_flock.assert_called_once_with(self.fileno, fcntl.LOCK_EX)
        mock_chown.assert_has_calls(
            [
                mock.call(FILE_PATH, user=FILE_OWNER),
                mock.call(FILE_PATH, group=FILE_GROUP),
            ]
        )
        self.assertEqual(mock_chown.call_count, 2)
        mock_chmod.assert_called_once_with(self.fileno, FILE_PERMISSIONS)
        self.assertEqual(cm.exception.metadata, raw_file.metadata)
        self.assertEqual(cm.exception.action, RawFileError.ACTION_WRITE)
        self.assertEqual(cm.exception.reason, f"some error: '{FILE_PATH}'")


@mock.patch.object(RawFile, "_chmod")
@mock.patch.object(RawFile, "_chown")
class RawFileUpdate(TestCase):
    fileno = 123

    def setUp(self):
        self.tmp_file = get_tmp_file(mode="a+b")
        self.raw_file = RawFile(
            fixture_metadata(
                owner=None,
                group=None,
                permissions=None,
                file_path=self.tmp_file.name,
            )
        )
        self.raw_file_binary = RawFile(
            fixture_metadata(
                owner=None,
                group=None,
                permissions=None,
                file_path=self.tmp_file.name,
                binary=True,
            )
        )

    def tearDown(self):
        self.tmp_file.close()

    def assert_success(self, raw_file, modify_fn, mock_chmod, mock_chown):
        orig_data = b"original data"
        write_data_to_tmpfile(orig_data, self.tmp_file)

        with raw_file.update() as io_buffer:
            with self.assertRaises(OSError):
                fcntl.flock(self.tmp_file, fcntl.LOCK_SH | fcntl.LOCK_NB)
            self.assertEqual(orig_data, io_buffer.getvalue())
            modify_fn(io_buffer)
            new_data = io_buffer.getvalue()
        self.tmp_file.seek(0)
        self.assertEqual(self.tmp_file.read(), new_data)
        mock_chmod.assert_called_once()
        mock_chown.assert_called_once()

    @staticmethod
    def _modify(io_buffer):
        io_buffer.seek(0)
        io_buffer.truncate()
        io_buffer.write(b"new")

    @staticmethod
    def _modify_no_truncate(io_buffer):
        io_buffer.write(b"new")

    def test_success(self, mock_chown, mock_chmod):
        self.assert_success(
            self.raw_file,
            self._modify,
            mock_chmod,
            mock_chown,
        )

    def test_success_no_truncate(self, mock_chown, mock_chmod):
        self.assert_success(
            self.raw_file,
            self._modify_no_truncate,
            mock_chmod,
            mock_chown,
        )

    def test_success_binary(self, mock_chown, mock_chmod):
        self.assert_success(
            self.raw_file_binary,
            self._modify,
            mock_chmod,
            mock_chown,
        )

    def test_success_no_truncate_binary(self, mock_chown, mock_chmod):
        self.assert_success(
            self.raw_file_binary,
            self._modify_no_truncate,
            mock_chmod,
            mock_chown,
        )

    def test_non_existing_file(self, mock_chown, mock_chmod):
        tmp_dir = get_tmp_dir()
        file_path = os.path.join(tmp_dir.name, "tmp_file")
        raw_file = RawFile(
            fixture_metadata(
                owner=None,
                group=None,
                permissions=None,
                file_path=file_path,
            )
        )
        with tmp_dir:
            with raw_file.update() as io_buffer:
                try:
                    # pylint: disable=consider-using-with
                    file_obj = open(file_path)
                except OSError:
                    self.fail("Unable to open file")
                with self.assertRaises(OSError):
                    fcntl.flock(file_obj, fcntl.LOCK_SH | fcntl.LOCK_NB)
                file_obj.close()
                self.assertEqual(b"", io_buffer.getvalue())
                self._modify_no_truncate(io_buffer)
                new_data = io_buffer.getvalue()
            with open(file_path, "rb") as file_obj:
                self.assertEqual(new_data, file_obj.read())
        mock_chmod.assert_called_once()
        mock_chown.assert_called_once()

    def test_open_error(self, mock_chown, mock_chmod):
        mock_open = mock.MagicMock()
        mock_open.side_effect = OSError()
        with patch_file("open", mock_open):
            with self.assertRaises(RawFileError):
                with self.raw_file.update():
                    self.fail("should not get here")
        mock_chmod.assert_not_called()
        mock_chown.assert_not_called()

    @patch_file("fcntl.flock")
    def test_read_error(self, mock_flock, mock_chown, mock_chmod):
        mock_open = mock.MagicMock()
        file_mock = mock_open.return_value.__enter__.return_value
        file_mock.fileno.return_value = self.fileno
        file_mock.read.side_effect = OSError()
        with patch_file("open", mock_open):
            with self.assertRaises(RawFileError):
                with self.raw_file.update():
                    self.fail("should not get here")
        mock_open.return_value.__enter__.return_value.read.assert_called_once_with()
        mock_flock.assert_called_once_with(self.fileno, fcntl.LOCK_EX)
        mock_chmod.assert_not_called()
        mock_chown.assert_not_called()

    @patch_file("fcntl.flock")
    def test_write_error(self, mock_flock, mock_chown, mock_chmod):
        orig_data = "original data"
        new_data = "new data"
        mock_open = mock.MagicMock()
        file_mock = mock_open.return_value.__enter__.return_value
        file_mock.fileno.return_value = self.fileno
        file_mock.read.return_value = orig_data
        file_mock.write.side_effect = OSError()
        with patch_file("open", mock_open):
            with self.assertRaises(RawFileError):
                with self.raw_file.update() as io_buffer:
                    self.assertEqual(
                        orig_data.encode("utf-8"), io_buffer.getvalue()
                    )
                    io_buffer.seek(0)
                    io_buffer.truncate()
                    io_buffer.write(new_data.encode("utf-8"))
                self.fail("should not get here")
        file_mock.read.assert_called_once_with()
        file_mock.write.assert_called_once_with(new_data)
        mock_flock.assert_called_once_with(self.fileno, fcntl.LOCK_EX)
        mock_chmod.assert_called_once_with(file_mock)
        mock_chown.assert_called_once_with()


@patch_file("os.remove")
class RawFileRemove(TestCase):
    # pylint: disable=no-self-use
    def test_success(self, mock_remove):
        RawFile(fixture_metadata()).remove()
        mock_remove.assert_called_once_with(FILE_PATH)

    def test_file_not_found(self, mock_remove):
        mock_remove.side_effect = OSError(errno.ENOENT, "some error", FILE_PATH)
        RawFile(fixture_metadata()).remove(fail_if_file_not_found=False)
        mock_remove.assert_called_once_with(FILE_PATH)

    def test_file_not_found_fail(self, mock_remove):
        raw_file = RawFile(fixture_metadata())
        mock_remove.side_effect = OSError(errno.ENOENT, "some error", FILE_PATH)
        with self.assertRaises(RawFileError) as cm:
            raw_file.remove(fail_if_file_not_found=True)
        mock_remove.assert_called_once_with(FILE_PATH)
        self.assertEqual(cm.exception.metadata, raw_file.metadata)
        self.assertEqual(cm.exception.action, RawFileError.ACTION_REMOVE)
        self.assertEqual(cm.exception.reason, f"some error: '{FILE_PATH}'")

    def test_cannot_remove(self, mock_remove):
        raw_file = RawFile(fixture_metadata())
        mock_remove.side_effect = OSError(1, "some error", FILE_PATH)
        with self.assertRaises(RawFileError) as cm:
            raw_file.remove(fail_if_file_not_found=True)
        mock_remove.assert_called_once_with(FILE_PATH)
        self.assertEqual(cm.exception.metadata, raw_file.metadata)
        self.assertEqual(cm.exception.action, RawFileError.ACTION_REMOVE)
        self.assertEqual(cm.exception.reason, f"some error: '{FILE_PATH}'")



from unittest import (
    TestCase,
    mock,
)

from pcs.common import reports
from pcs.lib import resource_agent as ra

from pcs_test.tools import fixture
from pcs_test.tools.command_env import get_env_tools


@mock.patch("pcs.lib.resource_agent.facade.ocf_unified_to_pcs")
class ResourceAgentFacade(TestCase):
    # Methods returning validators are not tested here. They are tested
    # indirectly in tests for methods calling them.

    @staticmethod
    def _fixture_metadata(shortdesc, parameters):
        return ra.ResourceAgentMetadata(
            ra.ResourceAgentName("standard", "provider", "type"),
            agent_exists=True,
            ocf_version=ra.const.OCF_1_0,
            shortdesc=shortdesc,
            longdesc=None,
            parameters=parameters,
            actions=[],
        )

    @staticmethod
    def _fixture_parameter(name):
        return ra.ResourceAgentParameter(
            name,
            shortdesc=None,
            longdesc=None,
            type="string",
            default=None,
            enum_values=None,
            required=False,
            advanced=False,
            deprecated=False,
            deprecated_by=[],
            deprecated_desc=None,
            unique_group=None,
            reloadable=False,
        )

    def test_metadata_transformation(self, mock_ocf_transform):
        metadata_in = self._fixture_metadata("raw metadata", [])
        metadata_out = self._fixture_metadata("transformed metadata", [])
        mock_ocf_transform.return_value = metadata_out

        facade = ra.ResourceAgentFacade(metadata_in, [])
        self.assertEqual(facade.metadata, metadata_out)
        mock_ocf_transform.assert_called_once_with(metadata_in)

        # transformation is cached
        mock_ocf_transform.reset_mock()
        self.assertEqual(facade.metadata, metadata_out)
        mock_ocf_transform.assert_not_called()

    def test_add_additional_params(self, mock_ocf_transform):
        metadata_in = self._fixture_metadata(
            "raw metadata", [self._fixture_parameter("param1")]
        )
        metadata_transformed = self._fixture_metadata(
            "transformed metadata", [self._fixture_parameter("param1")]
        )
        metadata_out = self._fixture_metadata(
            "transformed metadata",
            [
                self._fixture_parameter("param1"),
                self._fixture_parameter("param2"),
            ],
        )
        mock_ocf_transform.return_value = metadata_transformed

        facade = ra.ResourceAgentFacade(
            metadata_in, [self._fixture_parameter("param2")]
        )
        self.assertEqual(facade.metadata, metadata_out)
        mock_ocf_transform.assert_called_once_with(metadata_in)

        # transformation is cached
        mock_ocf_transform.reset_mock()
        self.assertEqual(facade.metadata, metadata_out)
        mock_ocf_transform.assert_not_called()


class ResourceAgentFacadeFactory(TestCase):
    _fixture_agent_xml = """
        <resource-agent name="agent">
            <parameters>
                <parameter name="agent-param"/>
            </parameters>
        </resource-agent>
    """
    _fixture_fenced_xml = """
        <resource-agent name="pacemaker-fenced">
            <parameters>
                <parameter name="fenced-param"/>
            </parameters>
        </resource-agent>
    """

    def setUp(self):
        self.env_assist, self.config = get_env_tools(test_case=self)

    def test_void(self):
        name = ra.ResourceAgentName("service", None, "daemon")
        env = self.env_assist.get_env()
        facade = ra.ResourceAgentFacadeFactory(
            env.cmd_runner(), env.report_processor
        ).void_facade_from_parsed_name(name)
        self.assertEqual(facade.metadata.name, name)
        self.assertFalse(facade.metadata.agent_exists)

    def test_facade(self):
        name = ra.ResourceAgentName("service", None, "daemon")
        self.config.runner.pcmk.load_agent(
            agent_name="service:daemon", stdout=self._fixture_agent_xml
        )

        env = self.env_assist.get_env()
        facade = ra.ResourceAgentFacadeFactory(
            env.cmd_runner(), env.report_processor
        ).facade_from_parsed_name(name)
        self.assertEqual(facade.metadata.name, name)
        self.assertTrue(facade.metadata.agent_exists)

    def test_facade_pacemaker(self):
        daemon_name = "pacemaker-fenced"
        fake_agent_name = ra.ResourceAgentName(
            ra.const.FAKE_AGENT_STANDARD, None, daemon_name
        )
        self.config.runner.pcmk.load_fake_agent_metadata(
            agent_name=daemon_name, stdout=self._fixture_fenced_xml
        )
        env = self.env_assist.get_env()
        facade = ra.ResourceAgentFacadeFactory(
            env.cmd_runner(), env.report_processor
        ).facade_from_pacemaker_daemon_name(daemon_name)
        self.assertEqual(facade.metadata.name, fake_agent_name)
        self.assertTrue(facade.metadata.agent_exists)
        self.assertEqual(
            [param.name for param in facade.metadata.parameters],
            ["fenced-param"],
        )

    def test_facade_missing_agent(self):
        name = ra.ResourceAgentName("service", None, "daemon")
        self.config.runner.pcmk.load_agent(
            agent_name="service:daemon", agent_is_missing=True
        )

        env = self.env_assist.get_env()
        with self.assertRaises(ra.UnableToGetAgentMetadata) as cm:
            ra.ResourceAgentFacadeFactory(
                env.cmd_runner(), env.report_processor
            ).facade_from_parsed_name(name)
        self.assertEqual(cm.exception.agent_name, name.full_name)

    def test_facade_pacemaker_unsupported_agent(self):
        daemon_name = "unsupported"
        env = self.env_assist.get_env()
        with self.assertRaises(ra.UnableToGetAgentMetadata) as cm:
            ra.ResourceAgentFacadeFactory(
                env.cmd_runner(), env.report_processor
            ).facade_from_pacemaker_daemon_name(daemon_name)
        self.assertEqual(cm.exception.agent_name, daemon_name)
        self.assertEqual(cm.exception.message, "Unknown agent")

    def test_facade_pacemaker_load_error(self):
        daemon_name = "pacemaker-fenced"
        self.config.runner.pcmk.load_fake_agent_metadata(
            agent_name=daemon_name, stdout="", stderr="error", returncode=1
        )
        env = self.env_assist.get_env()
        with self.assertRaises(ra.UnableToGetAgentMetadata) as cm:
            ra.ResourceAgentFacadeFactory(
                env.cmd_runner(), env.report_processor
            ).facade_from_pacemaker_daemon_name(daemon_name)
        self.assertEqual(cm.exception.agent_name, daemon_name)
        self.assertEqual(cm.exception.message, "error")

    def test_void_load_and_cache_fenced_for_stonith(self):
        name1 = ra.ResourceAgentName("stonith", None, "fence_xvm")
        name2 = ra.ResourceAgentName("stonith", None, "fence_virt")
        self.config.runner.pcmk.load_fake_agent_metadata(
            stdout=self._fixture_fenced_xml
        )

        env = self.env_assist.get_env()
        factory = ra.ResourceAgentFacadeFactory(
            env.cmd_runner(), env.report_processor
        )
        facade1 = factory.void_facade_from_parsed_name(name1)
        facade2 = factory.void_facade_from_parsed_name(name2)

        self.assertEqual(facade1.metadata.name, name1)
        self.assertEqual(facade2.metadata.name, name2)
        self.assertFalse(facade1.metadata.agent_exists)
        self.assertFalse(facade2.metadata.agent_exists)
        self.assertEqual(
            [param.name for param in facade1.metadata.parameters],
            ["fenced-param"],
        )
        self.assertEqual(
            [param.name for param in facade2.metadata.parameters],
            ["fenced-param"],
        )

    def test_facade_load_and_cache_fenced_for_stonith(self):
        name1 = ra.ResourceAgentName("stonith", None, "fence_xvm")
        name2 = ra.ResourceAgentName("stonith", None, "fence_virt")
        self.config.runner.pcmk.load_agent(
            agent_name="stonith:fence_xvm",
            stdout=self._fixture_agent_xml,
            name="runner.pcmk.load_agent.xvm",
        )
        self.config.runner.pcmk.load_fake_agent_metadata(
            stdout=self._fixture_fenced_xml
        )
        self.config.runner.pcmk.load_agent(
            agent_name="stonith:fence_virt",
            stdout=self._fixture_agent_xml,
            name="runner.pcmk.load_agent.virt",
        )

        env = self.env_assist.get_env()
        factory = ra.ResourceAgentFacadeFactory(
            env.cmd_runner(), env.report_processor
        )
        facade1 = factory.facade_from_parsed_name(name1)
        facade2 = factory.facade_from_parsed_name(name2)

        self.assertEqual(facade1.metadata.name, name1)
        self.assertEqual(facade2.metadata.name, name2)
        self.assertTrue(facade1.metadata.agent_exists)
        self.assertTrue(facade2.metadata.agent_exists)
        self.assertEqual(
            [param.name for param in facade1.metadata.parameters],
            ["agent-param", "fenced-param"],
        )
        self.assertEqual(
            [param.name for param in facade2.metadata.parameters],
            ["agent-param", "fenced-param"],
        )

    def test_void_load_and_cache_fenced_for_stonith_failure(self):
        name1 = ra.ResourceAgentName("stonith", None, "fence_xvm")
        name2 = ra.ResourceAgentName("stonith", None, "fence_virt")
        self.config.runner.pcmk.load_fake_agent_metadata(
            stdout="", stderr="fenced failure", returncode=1
        )

        env = self.env_assist.get_env()
        factory = ra.ResourceAgentFacadeFactory(
            env.cmd_runner(), env.report_processor
        )
        facade1 = factory.void_facade_from_parsed_name(name1)
        facade2 = factory.void_facade_from_parsed_name(name2)

        self.env_assist.assert_reports(
            [
                fixture.warn(
                    reports.codes.UNABLE_TO_GET_AGENT_METADATA,
                    agent="pacemaker-fenced",
                    reason="fenced failure",
                )
            ]
        )
        self.assertEqual(facade1.metadata.name, name1)
        self.assertEqual(facade2.metadata.name, name2)
        self.assertFalse(facade1.metadata.agent_exists)
        self.assertFalse(facade2.metadata.agent_exists)
        self.assertEqual(
            [param.name for param in facade1.metadata.parameters],
            [],
        )
        self.assertEqual(
            [param.name for param in facade2.metadata.parameters],
            [],
        )

    def test_facade_load_and_cache_fenced_for_stonith_failure(self):
        name1 = ra.ResourceAgentName("stonith", None, "fence_xvm")
        name2 = ra.ResourceAgentName("stonith", None, "fence_virt")
        self.config.runner.pcmk.load_agent(
            agent_name="stonith:fence_xvm",
            stdout=self._fixture_agent_xml,
            name="runner.pcmk.load_agent.xvm",
        )
        self.config.runner.pcmk.load_fake_agent_metadata(
            stdout="", stderr="fenced failure", returncode=1
        )
        self.config.runner.pcmk.load_agent(
            agent_name="stonith:fence_virt",
            stdout=self._fixture_agent_xml,
            name="runner.pcmk.load_agent.virt",
        )

        env = self.env_assist.get_env()
        factory = ra.ResourceAgentFacadeFactory(
            env.cmd_runner(), env.report_processor
        )
        facade1 = factory.facade_from_parsed_name(name1)
        facade2 = factory.facade_from_parsed_name(name2)

        self.env_assist.assert_reports(
            [
                fixture.warn(
                    reports.codes.UNABLE_TO_GET_AGENT_METADATA,
                    agent="pacemaker-fenced",
                    reason="fenced failure",
                )
            ]
        )
        self.assertEqual(facade1.metadata.name, name1)
        self.assertEqual(facade2.metadata.name, name2)
        self.assertTrue(facade1.metadata.agent_exists)
        self.assertTrue(facade2.metadata.agent_exists)
        self.assertEqual(
            [param.name for param in facade1.metadata.parameters],
            ["agent-param"],
        )
        self.assertEqual(
            [param.name for param in facade2.metadata.parameters],
            ["agent-param"],
        )

    def test_void_add_trace_for_ocf(self):
        env = self.env_assist.get_env()
        for provider in ["heartbeat", "pacemaker"]:
            with self.subTest(provider=provider):
                name = ra.ResourceAgentName("ocf", provider, "Dummy")
                facade = ra.ResourceAgentFacadeFactory(
                    env.cmd_runner(), env.report_processor
                ).void_facade_from_parsed_name(name)
                self.assertEqual(facade.metadata.name, name)
                self.assertFalse(facade.metadata.agent_exists)
                self.assertEqual(
                    [param.name for param in facade.metadata.parameters],
                    ["trace_ra", "trace_file"],
                )

    def test_facade_add_trace_for_ocf(self):
        self.config.runner.pcmk.load_agent(
            agent_name="ocf:heartbeat:Dummy",
            stdout=self._fixture_agent_xml,
            name="runner.pcmk.load_agent.heartbeat",
        )
        self.config.runner.pcmk.load_agent(
            agent_name="ocf:pacemaker:Dummy",
            stdout=self._fixture_agent_xml,
            name="runner.pcmk.load_agent.pacemaker",
        )
        env = self.env_assist.get_env()

        for provider in ["heartbeat", "pacemaker"]:
            with self.subTest(provider=provider):
                name = ra.ResourceAgentName("ocf", provider, "Dummy")
                facade = ra.ResourceAgentFacadeFactory(
                    env.cmd_runner(), env.report_processor
                ).facade_from_parsed_name(name)
                self.assertEqual(facade.metadata.name, name)
                self.assertTrue(facade.metadata.agent_exists)
                self.assertEqual(
                    [param.name for param in facade.metadata.parameters],
                    ["agent-param", "trace_ra", "trace_file"],
                )

    def test_void_not_add_trace_for_ocf(self):
        name = ra.ResourceAgentName("ocf", "openstack", "Dummy")
        env = self.env_assist.get_env()
        facade = ra.ResourceAgentFacadeFactory(
            env.cmd_runner(), env.report_processor
        ).void_facade_from_parsed_name(name)
        self.assertEqual(facade.metadata.name, name)
        self.assertFalse(facade.metadata.agent_exists)
        self.assertEqual(
            [param.name for param in facade.metadata.parameters],
            [],
        )

    def test_facade_not_add_trace_for_ocf(self):
        name = ra.ResourceAgentName("ocf", "openstack", "Dummy")
        self.config.runner.pcmk.load_agent(
            agent_name="ocf:openstack:Dummy", stdout=self._fixture_agent_xml
        )
        env = self.env_assist.get_env()

        facade = ra.ResourceAgentFacadeFactory(
            env.cmd_runner(), env.report_processor
        ).facade_from_parsed_name(name)
        self.assertEqual(facade.metadata.name, name)
        self.assertTrue(facade.metadata.agent_exists)
        self.assertEqual(
            [param.name for param in facade.metadata.parameters],
            ["agent-param"],
        )



from typing import (
    Any,
    Collection,
    Container,
    Mapping,
    Optional,
)

from lxml.etree import _Element

from pcs.common import reports
from pcs.common.pacemaker.defaults import CibDefaultsDto
from pcs.common.pacemaker.nvset import CibNvsetDto
from pcs.common.reports.item import ReportItem
from pcs.common.tools import Version
from pcs.common.types import StringCollection
from pcs.lib.cib import (
    nvpair_multi,
    sections,
)
from pcs.lib.cib.rule import (
    RuleInEffectEval,
    RuleInEffectEvalDummy,
    RuleInEffectEvalOneByOne,
    RuleParseError,
    has_node_attr_expr_with_type_integer,
    has_rsc_or_op_expression,
    parse_rule,
)
from pcs.lib.cib.tools import (
    IdProvider,
    get_pacemaker_version_by_which_cib_was_validated,
)
from pcs.lib.env import LibraryEnvironment
from pcs.lib.errors import LibraryError
from pcs.lib.external import CommandRunner
from pcs.lib.pacemaker.live import has_rule_in_effect_status_tool


def resource_defaults_create(
    env: LibraryEnvironment,
    nvpairs: Mapping[str, str],
    nvset_options: Mapping[str, str],
    nvset_rule: Optional[str] = None,
    force_flags: Collection[reports.types.ForceCode] = (),
) -> None:
    """
    Create new resource defaults nvset

    env --
    nvpairs -- name-value pairs to be put into the new nvset
    nvset_options -- additional attributes of the created nvset
    nvset_rule -- optional rule describing when the created nvset applies
    force_flags -- list of flags codes
    """
    return _defaults_create(
        env,
        sections.RSC_DEFAULTS,
        dict(
            rule_allows_rsc_expr=True,
            rule_allows_op_expr=False,
            rule_allows_node_attr_expr=False,
        ),
        nvpairs,
        nvset_options,
        nvset_rule=nvset_rule,
        force_flags=force_flags,
    )


def operation_defaults_create(
    env: LibraryEnvironment,
    nvpairs: Mapping[str, str],
    nvset_options: Mapping[str, str],
    nvset_rule: Optional[str] = None,
    force_flags: Collection[reports.types.ForceCode] = (),
) -> None:
    """
    Create new operation defaults nvset

    env --
    nvpairs -- name-value pairs to be put into the new nvset
    nvset_options -- additional attributes of the created nvset
    nvset_rule -- optional rule describing when the created nvset applies
    force_flags -- list of flags codes
    """
    return _defaults_create(
        env,
        sections.OP_DEFAULTS,
        dict(
            rule_allows_rsc_expr=True,
            rule_allows_op_expr=True,
            rule_allows_node_attr_expr=True,
        ),
        nvpairs,
        nvset_options,
        nvset_rule=nvset_rule,
        force_flags=force_flags,
    )


def _defaults_create(
    env: LibraryEnvironment,
    cib_section_name: str,
    validator_options: Mapping[str, Any],
    nvpairs: Mapping[str, str],
    nvset_options: Mapping[str, str],
    nvset_rule: Optional[str] = None,
    force_flags: Container[reports.types.ForceCode] = (),
) -> None:
    required_cib_version = None
    nice_to_have_cib_version = None
    if nvset_rule:
        # Parse the rule to see if we need to upgrade CIB schema. All errors
        # would be properly reported by a validator called bellow, so we can
        # safely ignore them here.
        try:
            rule_tree = parse_rule(nvset_rule)
            if has_rsc_or_op_expression(rule_tree):
                required_cib_version = Version(3, 4, 0)
            if has_node_attr_expr_with_type_integer(rule_tree):
                nice_to_have_cib_version = Version(3, 5, 0)
        except RuleParseError:
            pass

    cib = env.get_cib(
        minimal_version=required_cib_version,
        nice_to_have_version=nice_to_have_cib_version,
    )
    id_provider = IdProvider(cib)

    validator = nvpair_multi.ValidateNvsetAppendNew(
        id_provider,
        nvpairs,
        nvset_options,
        nvset_rule=nvset_rule,
        **validator_options,
    )
    if env.report_processor.report_list(
        validator.validate(force_options=reports.codes.FORCE in force_flags)
    ).has_errors:
        raise LibraryError()

    nvpair_multi.nvset_append_new(
        sections.get(cib, cib_section_name),
        id_provider,
        get_pacemaker_version_by_which_cib_was_validated(cib),
        nvpair_multi.NVSET_META,
        nvpairs,
        nvset_options,
        nvset_rule=validator.get_parsed_rule(),
    )

    env.report_processor.report(
        ReportItem.warning(reports.messages.DefaultsCanBeOverridden())
    )
    env.push_cib()


def resource_defaults_config(
    env: LibraryEnvironment, evaluate_expired: bool
) -> CibDefaultsDto:
    """
    List all resource defaults nvsets

    env --
    evaluate_expired -- also evaluate whether rules are expired or in effect
    """
    cib = env.get_cib()
    rule_evaluator = _get_rule_evaluator(
        cib, env.cmd_runner(), env.report_processor, evaluate_expired
    )

    def get_config(tag: nvpair_multi.NvsetTag) -> list[CibNvsetDto]:
        return _defaults_config(
            cib,
            tag,
            sections.RSC_DEFAULTS,
            rule_evaluator,
        )

    return CibDefaultsDto(
        instance_attributes=get_config(nvpair_multi.NVSET_INSTANCE),
        meta_attributes=get_config(nvpair_multi.NVSET_META),
    )


def operation_defaults_config(
    env: LibraryEnvironment, evaluate_expired: bool
) -> CibDefaultsDto:
    """
    List all operation defaults nvsets

    env --
    evaluate_expired -- also evaluate whether rules are expired or in effect
    """
    cib = env.get_cib()
    rule_evaluator = _get_rule_evaluator(
        cib, env.cmd_runner(), env.report_processor, evaluate_expired
    )

    def get_config(tag: nvpair_multi.NvsetTag) -> list[CibNvsetDto]:
        return _defaults_config(
            cib,
            tag,
            sections.OP_DEFAULTS,
            rule_evaluator,
        )

    return CibDefaultsDto(
        instance_attributes=get_config(nvpair_multi.NVSET_INSTANCE),
        meta_attributes=get_config(nvpair_multi.NVSET_META),
    )


def _get_rule_evaluator(
    cib: _Element,
    runner: CommandRunner,
    report_processor: reports.ReportProcessor,
    evaluate_expired: bool,
) -> RuleInEffectEval:
    if evaluate_expired:
        if has_rule_in_effect_status_tool():
            return RuleInEffectEvalOneByOne(cib, runner)
        report_processor.report(
            ReportItem.warning(
                reports.messages.RuleInEffectStatusDetectionNotSupported()
            )
        )
    return RuleInEffectEvalDummy()


def _defaults_config(
    cib: _Element,
    nvset_tag: nvpair_multi.NvsetTag,
    cib_section_name: str,
    rule_evaluator: RuleInEffectEval,
) -> list[CibNvsetDto]:
    return [
        nvpair_multi.nvset_element_to_dto(nvset_el, rule_evaluator)
        for nvset_el in nvpair_multi.find_nvsets(
            sections.get(cib, cib_section_name), nvset_tag
        )
    ]


def resource_defaults_remove(
    env: LibraryEnvironment, nvset_id_list: StringCollection
) -> None:
    """
    Remove specified resource defaults nvsets

    env --
    nvset_id_list -- nvset IDs to be removed
    """
    return _defaults_remove(env, sections.RSC_DEFAULTS, nvset_id_list)


def operation_defaults_remove(
    env: LibraryEnvironment, nvset_id_list: StringCollection
) -> None:
    """
    Remove specified operation defaults nvsets

    env --
    nvset_id_list -- nvset IDs to be removed
    """
    return _defaults_remove(env, sections.OP_DEFAULTS, nvset_id_list)


def _defaults_remove(
    env: LibraryEnvironment,
    cib_section_name: str,
    nvset_id_list: StringCollection,
) -> None:
    if not nvset_id_list:
        return
    nvset_elements, report_list = nvpair_multi.find_nvsets_by_ids(
        sections.get(env.get_cib(), cib_section_name), nvset_id_list
    )
    if env.report_processor.report_list(report_list).has_errors:
        raise LibraryError()
    nvpair_multi.nvset_remove(nvset_elements)
    env.push_cib()


def resource_defaults_update(
    env: LibraryEnvironment,
    nvset_id: Optional[str],
    nvpairs: Mapping[str, str],
) -> None:
    """
    Update specified resource defaults nvset

    env --
    nvset_id -- nvset ID to be updated; if None, update an existing nvset if
        there is only one
    nvpairs -- name-value pairs to be put into the nvset
    """
    return _defaults_update(
        env,
        sections.RSC_DEFAULTS,
        nvset_id,
        nvpairs,
        reports.const.PCS_COMMAND_RESOURCE_DEFAULTS_UPDATE,
    )


def operation_defaults_update(
    env: LibraryEnvironment,
    nvset_id: Optional[str],
    nvpairs: Mapping[str, str],
) -> None:
    """
    Update specified operation defaults nvset

    env --
    nvset_id -- nvset ID to be updated; if None, update an existing nvset if
        there is only one
    nvpairs -- name-value pairs to be put into the nvset
    """
    return _defaults_update(
        env,
        sections.OP_DEFAULTS,
        nvset_id,
        nvpairs,
        reports.const.PCS_COMMAND_OPERATION_DEFAULTS_UPDATE,
    )


def _defaults_update(
    env: LibraryEnvironment,
    cib_section_name: str,
    nvset_id: Optional[str],
    nvpairs: Mapping[str, str],
    pcs_command: reports.types.PcsCommand,
) -> None:
    cib = env.get_cib()
    id_provider = IdProvider(cib)

    if nvset_id is None:
        # Backward compatibility code to support an old use case where no id
        # was requested and provided and the first meta_attributes nvset was
        # created / updated. However, we check that there is only one nvset
        # present in the CIB to prevent breaking the configuration with
        # multiple nvsets in place.

        # This is to be supported as it provides means of easily managing
        # defaults if only one set of defaults is needed.

        # TODO move this to a separate lib command.

        if not nvpairs:
            return

        # Do not create new defaults element if we are only removing values
        # from it.
        only_removing = True
        for value in nvpairs.values():
            if value != "":
                only_removing = False
                break
        if only_removing and not sections.exists(cib, cib_section_name):
            env.report_processor.report(
                ReportItem.warning(reports.messages.DefaultsCanBeOverridden())
            )
            return

        nvset_elements = nvpair_multi.find_nvsets(
            sections.get(cib, cib_section_name), nvpair_multi.NVSET_META
        )
        if len(nvset_elements) > 1:
            env.report_processor.report(
                reports.item.ReportItem.error(
                    reports.messages.CibNvsetAmbiguousProvideNvsetId(
                        pcs_command
                    )
                )
            )
            raise LibraryError()
        env.report_processor.report(
            ReportItem.warning(reports.messages.DefaultsCanBeOverridden())
        )
        if len(nvset_elements) == 1:
            nvpair_multi.nvset_update(nvset_elements[0], id_provider, nvpairs)
        elif only_removing:
            # do not create new nvset if there is none and we are only removing
            # nvpairs
            return
        else:
            nvpair_multi.nvset_append_new(
                sections.get(cib, cib_section_name),
                id_provider,
                get_pacemaker_version_by_which_cib_was_validated(cib),
                nvpair_multi.NVSET_META,
                nvpairs,
                {},
            )
        env.push_cib()
        return

    nvset_elements, report_list = nvpair_multi.find_nvsets_by_ids(
        sections.get(cib, cib_section_name), [nvset_id]
    )
    if env.report_processor.report_list(report_list).has_errors:
        raise LibraryError()

    nvpair_multi.nvset_update(nvset_elements[0], id_provider, nvpairs)
    env.report_processor.report(
        ReportItem.warning(reports.messages.DefaultsCanBeOverridden())
    )
    env.push_cib()



from typing import Optional


class ManageServiceError(Exception):
    # pylint: disable=super-init-not-called
    def __init__(
        self,
        service: str,
        message: str,
        instance: Optional[str] = None,
    ):
        self.service = service
        self.message = message
        self.instance = instance


class DisableServiceError(ManageServiceError):
    pass


class EnableServiceError(ManageServiceError):
    pass


class StartServiceError(ManageServiceError):
    pass


class StopServiceError(ManageServiceError):
    pass



from pcs.cli.common import parse_args
from pcs.cli.common.errors import CmdLineInputError


def separate_tail_option_candidates(arg_list):
    for i, arg in enumerate(arg_list):
        if "=" in arg:
            return arg_list[:i], arg_list[i:]

    return arg_list, []


def parse_add(arg_list):
    info, option_candidates = separate_tail_option_candidates(arg_list)

    if not info:
        raise CmdLineInputError("Ticket not specified")

    ticket, resource_specification = info[0], info[1:]

    if len(resource_specification) not in (1, 2):
        raise CmdLineInputError(
            "invalid resource specification: '{0}'".format(
                " ".join(resource_specification)
            )
        )

    if len(resource_specification) == 2:
        resource_role, resource_id = resource_specification
    else:
        resource_role = ""
        resource_id = resource_specification[0]

    return (
        ticket,
        resource_id,
        resource_role,
        parse_args.prepare_options(option_candidates),
    )



from unittest import TestCase

from pcs.daemon.app.api_v1 import API_V1_MAP
from pcs.daemon.async_tasks.worker.command_mapping import COMMAND_MAP


class ApiV1MapTest(TestCase):
    def test_all_commands_exist(self):
        missing_commands = set(API_V1_MAP.values()) - set(COMMAND_MAP.keys())
        self.assertEqual(
            0,
            len(missing_commands),
            f"Commands missing in COMMAND_MAP: {missing_commands}",
        )



from enum import Enum
from typing import (
    FrozenSet,
    Optional,
)

import dacite

from pcs.common import file_type_codes as code
from pcs.common import reports
from pcs.lib.file.json import JsonParser
from pcs.lib.interface.config import (
    ParserErrorException,
    ParserInterface,
)

from .types import ConfigV2


class ParserError(ParserErrorException):
    def __init__(self, msg: Optional[str] = None) -> None:
        super().__init__()
        self.msg = msg


class ParserV2(ParserInterface):
    @staticmethod
    def parse(raw_file_data: bytes) -> ConfigV2:
        data = JsonParser.parse(raw_file_data)
        try:
            if not isinstance(data, dict):
                raise ParserError("Invalid data: root element is not an object")
            try:
                version = data.pop("format_version")
            except KeyError as e:
                raise ParserError("'format_version' not defined") from e
            if version != 2:
                raise ParserError(f"Unsupported format version '{version}'")
            return dacite.from_dict(
                ConfigV2, data, config=dacite.Config(cast=[Enum, FrozenSet])
            )
        except dacite.DaciteError as e:
            raise ParserError(str(e)) from e
        except ValueError as e:
            raise ParserError(str(e)) from e

    @staticmethod
    def exception_to_report_list(
        exception: ParserErrorException,
        file_type_code: code.FileTypeCode,
        file_path: Optional[str],
        force_code: Optional[reports.types.ForceCode],
        is_forced_or_warning: bool,
    ) -> reports.ReportItemList:
        raise NotImplementedError()



import os
import sys

from pcs import utils
from pcs.cli.common.errors import CmdLineInputError
from pcs.cli.reports import process_library_reports
from pcs.cli.reports.output import error
from pcs.lib.errors import LibraryError
from pcs.lib.node import get_existing_nodes_names
from pcs.lib.pacemaker.live import get_cluster_status_dom
from pcs.lib.pacemaker.state import ClusterState
from pcs.pcsd import pcsd_status_cmd

# pylint: disable=too-many-branches, too-many-locals, too-many-statements


def full_status(lib, argv, modifiers):
    """
    Options:
      * --hide-inactive - hide inactive resources
      * --full - show full details, node attributes and failcount
      * -f - CIB file, crm_mon accepts CIB_file environment variable
      * --corosync_conf - file corocync.conf
      * --request-timeout - HTTP timeout for node authorization check
    """
    modifiers.ensure_only_supported(
        "--hide-inactive",
        "--full",
        "-f",
        "--corosync_conf",
        "--request-timeout",
    )
    if argv:
        raise CmdLineInputError()
    print(
        lib.status.full_cluster_status_plaintext(
            hide_inactive_resources=modifiers.get("--hide-inactive"),
            verbose=modifiers.get("--full"),
        )
    )


# Parse crm_mon for status
def nodes_status(lib, argv, modifiers):
    """
    Options:
      * -f - CIB file - for config subcommand and not for both or corosync
      * --corosync_conf - only for config subcommand

    NOTE: modifiers check is in subcommand
    """
    del lib
    if len(argv) == 1 and (argv[0] == "config"):
        modifiers.ensure_only_supported("-f", "--corosync_conf")
        if utils.hasCorosyncConf():
            corosync_nodes, report_list = get_existing_nodes_names(
                utils.get_corosync_conf_facade()
            )
            if report_list:
                process_library_reports(report_list)
        else:
            corosync_nodes = []
        try:
            pacemaker_nodes = sorted(
                [
                    node.attrs.name
                    for node in ClusterState(
                        get_cluster_status_dom(utils.cmd_runner())
                    ).node_section.nodes
                    if node.attrs.type != "remote"
                ]
            )
        except LibraryError as e:
            process_library_reports(e.args)
        print("Corosync Nodes:")
        if corosync_nodes:
            print(" " + " ".join(corosync_nodes))
        print("Pacemaker Nodes:")
        if pacemaker_nodes:
            print(" " + " ".join(pacemaker_nodes))

        return

    if len(argv) == 1 and (argv[0] == "corosync" or argv[0] == "both"):
        modifiers.ensure_only_supported()
        all_nodes, report_list = get_existing_nodes_names(
            utils.get_corosync_conf_facade()
        )
        if report_list:
            process_library_reports(report_list)
        online_nodes = utils.getCorosyncActiveNodes()
        offline_nodes = []
        for node in all_nodes:
            if node not in online_nodes:
                offline_nodes.append(node)

        online_nodes.sort()
        offline_nodes.sort()
        print("Corosync Nodes:")
        print(" ".join([" Online:"] + online_nodes))
        print(" ".join([" Offline:"] + offline_nodes))
        if argv[0] != "both":
            sys.exit(0)

    modifiers.ensure_only_supported("-f")
    info_dom = utils.getClusterState()

    nodes = info_dom.getElementsByTagName("nodes")
    if nodes.length == 0:
        utils.err("No nodes section found")

    onlinenodes = []
    offlinenodes = []
    standbynodes = []
    standbynodes_with_resources = []
    maintenancenodes = []
    remote_onlinenodes = []
    remote_offlinenodes = []
    remote_standbynodes = []
    remote_standbynodes_with_resources = []
    remote_maintenancenodes = []
    for node in nodes[0].getElementsByTagName("node"):
        node_name = node.getAttribute("name")
        node_remote = node.getAttribute("type") == "remote"
        if node.getAttribute("online") == "true":
            if node.getAttribute("standby") == "true":
                is_running_resources = (
                    node.getAttribute("resources_running") != "0"
                )
                if node_remote:
                    if is_running_resources:
                        remote_standbynodes_with_resources.append(node_name)
                    else:
                        remote_standbynodes.append(node_name)
                else:
                    if is_running_resources:
                        standbynodes_with_resources.append(node_name)
                    else:
                        standbynodes.append(node_name)
            if node.getAttribute("maintenance") == "true":
                if node_remote:
                    remote_maintenancenodes.append(node_name)
                else:
                    maintenancenodes.append(node_name)
            if (
                node.getAttribute("standby") == "false"
                and node.getAttribute("maintenance") == "false"
            ):
                if node_remote:
                    remote_onlinenodes.append(node_name)
                else:
                    onlinenodes.append(node_name)
        else:
            if node_remote:
                remote_offlinenodes.append(node_name)
            else:
                offlinenodes.append(node_name)

    print("Pacemaker Nodes:")
    print(" ".join([" Online:"] + onlinenodes))
    print(" ".join([" Standby:"] + standbynodes))
    print(
        " ".join(
            [" Standby with resource(s) running:"] + standbynodes_with_resources
        )
    )
    print(" ".join([" Maintenance:"] + maintenancenodes))
    print(" ".join([" Offline:"] + offlinenodes))

    print("Pacemaker Remote Nodes:")
    print(" ".join([" Online:"] + remote_onlinenodes))
    print(" ".join([" Standby:"] + remote_standbynodes))
    print(
        " ".join(
            [" Standby with resource(s) running:"]
            + remote_standbynodes_with_resources
        )
    )
    print(" ".join([" Maintenance:"] + remote_maintenancenodes))
    print(" ".join([" Offline:"] + remote_offlinenodes))


def cluster_status(lib, argv, modifiers):
    """
    Options:
      * -f - CIB file
      * --request-timeout - HTTP timeout for checking status of pcsd, no effect
        if -f is specified
    """
    modifiers.ensure_only_supported("-f", "--request-timeout")
    if argv:
        raise CmdLineInputError()
    (output, retval) = utils.run(["crm_mon", "-1", "-r"])

    if retval != 0:
        utils.err("cluster is not currently running on this node")

    first_empty_line = False
    print("Cluster Status:")
    for line in output.splitlines():
        if line == "":
            if first_empty_line:
                break
            first_empty_line = True
            continue
        print("", line)

    if not modifiers.is_specified("-f") and utils.hasCorosyncConf():
        print()
        print_pcsd_daemon_status(lib, modifiers)


def corosync_status(lib, argv, modifiers):
    """
    Options: no options
    """
    del lib
    modifiers.ensure_only_supported()
    if argv:
        raise CmdLineInputError()
    (output, retval) = utils.run(["corosync-quorumtool", "-l"])
    if retval != 0:
        utils.err("corosync not running")
    else:
        print(output.rstrip())


def xml_status(lib, argv, modifiers):
    """
    Options:
      * -f - CIB file
    """
    modifiers.ensure_only_supported("-f")
    if argv:
        raise CmdLineInputError()
    print(lib.status.pacemaker_status_xml().strip())


def print_pcsd_daemon_status(lib, modifiers):
    """
    Commandline options:
      * --request-timeout - HTTP timeout for node authorization check or when
        not running under root to call local pcsd
    """
    print("PCSD Status:")
    if os.getuid() == 0:
        pcsd_status_cmd(
            lib, [], modifiers.get_subset("--request-timeout"), dont_exit=True
        )
    else:
        err_msgs, exitcode, std_out, dummy_std_err = utils.call_local_pcsd(
            ["status", "pcsd"], []
        )
        if err_msgs:
            for msg in err_msgs:
                print(msg)
        if exitcode == 0:
            print(std_out)
        else:
            raise error("Unable to get PCSD status")



from dataclasses import dataclass
from typing import (
    Any,
    Callable,
    Mapping,
)

from pcs.lib.commands import (  # services,
    acl,
    alert,
    cluster,
    cluster_property,
    constraint,
    fencing_topology,
    node,
    qdevice,
    quorum,
    resource,
    resource_agent,
    sbd,
    scsi,
    status,
    stonith,
    stonith_agent,
)
from pcs.lib.permissions.config.types import PermissionAccessType as p


@dataclass(frozen=True)
class _Cmd:
    cmd: Callable[..., Any]
    required_permission: p


COMMAND_MAP: Mapping[str, _Cmd] = {
    "acl.add_permission": _Cmd(
        cmd=acl.add_permission,
        required_permission=p.GRANT,
    ),
    "acl.assign_role_to_group": _Cmd(
        cmd=acl.assign_role_to_group,
        required_permission=p.GRANT,
    ),
    "acl.assign_role_to_target": _Cmd(
        cmd=acl.assign_role_to_target,
        required_permission=p.GRANT,
    ),
    "acl.create_group": _Cmd(
        cmd=acl.create_group,
        required_permission=p.GRANT,
    ),
    "acl.create_role": _Cmd(
        cmd=acl.create_role,
        required_permission=p.GRANT,
    ),
    "acl.create_target": _Cmd(
        cmd=acl.create_target,
        required_permission=p.GRANT,
    ),
    "acl.remove_group": _Cmd(
        cmd=acl.remove_group,
        required_permission=p.GRANT,
    ),
    "acl.remove_permission": _Cmd(
        cmd=acl.remove_permission,
        required_permission=p.GRANT,
    ),
    "acl.remove_role": _Cmd(
        cmd=acl.remove_role,
        required_permission=p.GRANT,
    ),
    "acl.remove_target": _Cmd(
        cmd=acl.remove_target,
        required_permission=p.GRANT,
    ),
    "acl.unassign_role_from_group": _Cmd(
        cmd=acl.unassign_role_from_group,
        required_permission=p.GRANT,
    ),
    "acl.unassign_role_from_target": _Cmd(
        cmd=acl.unassign_role_from_target,
        required_permission=p.GRANT,
    ),
    "alert.add_recipient": _Cmd(
        cmd=alert.add_recipient,
        required_permission=p.WRITE,
    ),
    "alert.create_alert": _Cmd(
        cmd=alert.create_alert,
        required_permission=p.WRITE,
    ),
    "alert.remove_alert": _Cmd(
        cmd=alert.remove_alert,
        required_permission=p.WRITE,
    ),
    "alert.remove_recipient": _Cmd(
        cmd=alert.remove_recipient,
        required_permission=p.WRITE,
    ),
    "alert.update_alert": _Cmd(
        cmd=alert.update_alert,
        required_permission=p.WRITE,
    ),
    "alert.update_recipient": _Cmd(
        cmd=alert.update_recipient,
        required_permission=p.WRITE,
    ),
    "cluster.add_nodes": _Cmd(
        cmd=cluster.add_nodes,
        required_permission=p.FULL,
    ),
    "cluster.generate_cluster_uuid": _Cmd(
        cmd=cluster.generate_cluster_uuid,
        required_permission=p.SUPERUSER,
    ),
    "cluster.node_clear": _Cmd(
        cmd=cluster.node_clear,
        required_permission=p.WRITE,
    ),
    "cluster.remove_nodes": _Cmd(
        cmd=cluster.remove_nodes,
        required_permission=p.FULL,
    ),
    "cluster.setup": _Cmd(
        cmd=cluster.setup,
        required_permission=p.SUPERUSER,
    ),
    "cluster_property.set_properties": _Cmd(
        cmd=cluster_property.set_properties,
        required_permission=p.WRITE,
    ),
    "constraint.colocation.create_with_set": _Cmd(
        cmd=constraint.colocation.create_with_set,
        required_permission=p.WRITE,
    ),
    "constraint.order.create_with_set": _Cmd(
        cmd=constraint.order.create_with_set,
        required_permission=p.WRITE,
    ),
    "constraint.ticket.create": _Cmd(
        cmd=constraint.ticket.create,
        required_permission=p.WRITE,
    ),
    "constraint.ticket.create_with_set": _Cmd(
        cmd=constraint.ticket.create_with_set,
        required_permission=p.WRITE,
    ),
    "constraint.ticket.remove": _Cmd(
        cmd=constraint.ticket.remove,
        required_permission=p.WRITE,
    ),
    "fencing_topology.add_level": _Cmd(
        cmd=fencing_topology.add_level,
        required_permission=p.WRITE,
    ),
    "fencing_topology.remove_all_levels": _Cmd(
        cmd=fencing_topology.remove_all_levels,
        required_permission=p.WRITE,
    ),
    "fencing_topology.remove_levels_by_params": _Cmd(
        cmd=fencing_topology.remove_levels_by_params,
        required_permission=p.WRITE,
    ),
    "fencing_topology.verify": _Cmd(
        cmd=fencing_topology.verify,
        required_permission=p.WRITE,
    ),
    "node.maintenance_unmaintenance_all": _Cmd(
        cmd=node.maintenance_unmaintenance_all,
        required_permission=p.WRITE,
    ),
    "node.maintenance_unmaintenance_list": _Cmd(
        cmd=node.maintenance_unmaintenance_list,
        required_permission=p.WRITE,
    ),
    "node.standby_unstandby_all": _Cmd(
        cmd=node.standby_unstandby_all,
        required_permission=p.WRITE,
    ),
    "node.standby_unstandby_list": _Cmd(
        cmd=node.standby_unstandby_list,
        required_permission=p.WRITE,
    ),
    "qdevice.client_net_import_certificate": _Cmd(
        cmd=qdevice.client_net_import_certificate,
        required_permission=p.FULL,
    ),
    # TODO: make sure WRITE is right permission
    "qdevice.qdevice_net_sign_certificate_request": _Cmd(
        cmd=qdevice.qdevice_net_sign_certificate_request,
        required_permission=p.WRITE,
    ),
    "quorum.device_net_certificate_check_local": _Cmd(
        cmd=quorum.device_net_certificate_check_local,
        required_permission=p.READ,
    ),
    "quorum.device_net_certificate_setup_local": _Cmd(
        cmd=quorum.device_net_certificate_setup_local,
        required_permission=p.WRITE,
    ),
    # deprecated, API v1 compatibility
    "resource_agent.describe_agent": _Cmd(
        cmd=resource_agent.describe_agent,
        required_permission=p.READ,
    ),
    "resource_agent.get_agents_list": _Cmd(
        cmd=resource_agent.get_agents_list,
        required_permission=p.READ,
    ),
    "resource_agent.get_agent_metadata": _Cmd(
        cmd=resource_agent.get_agent_metadata,
        required_permission=p.READ,
    ),
    # deprecated, API v1 compatibility
    "resource_agent.list_agents": _Cmd(
        cmd=resource_agent.list_agents,
        required_permission=p.READ,
    ),
    # deprecated, API v1 compatibility
    "resource_agent.list_agents_for_standard_and_provider": _Cmd(
        cmd=resource_agent.list_agents_for_standard_and_provider,
        required_permission=p.READ,
    ),
    # deprecated, API v1 compatibility
    "resource_agent.list_ocf_providers": _Cmd(
        cmd=resource_agent.list_ocf_providers,
        required_permission=p.READ,
    ),
    # deprecated, API v1 compatibility
    "resource_agent.list_standards": _Cmd(
        cmd=resource_agent.list_standards,
        required_permission=p.READ,
    ),
    "resource.ban": _Cmd(
        cmd=resource.ban,
        required_permission=p.WRITE,
    ),
    "resource.create": _Cmd(
        cmd=resource.create,
        required_permission=p.WRITE,
    ),
    "resource.create_as_clone": _Cmd(
        cmd=resource.create_as_clone,
        required_permission=p.WRITE,
    ),
    "resource.create_in_group": _Cmd(
        cmd=resource.create_in_group,
        required_permission=p.WRITE,
    ),
    "resource.disable": _Cmd(
        cmd=resource.disable,
        required_permission=p.WRITE,
    ),
    "resource.disable_safe": _Cmd(
        cmd=resource.disable_safe,
        required_permission=p.WRITE,
    ),
    "resource.disable_simulate": _Cmd(
        cmd=resource.disable_simulate,
        required_permission=p.READ,
    ),
    "resource.enable": _Cmd(
        cmd=resource.enable,
        required_permission=p.WRITE,
    ),
    "resource.group_add": _Cmd(
        cmd=resource.group_add,
        required_permission=p.WRITE,
    ),
    "resource.manage": _Cmd(
        cmd=resource.manage,
        required_permission=p.WRITE,
    ),
    "resource.move": _Cmd(
        cmd=resource.move,
        required_permission=p.WRITE,
    ),
    "resource.move_autoclean": _Cmd(
        cmd=resource.move_autoclean,
        required_permission=p.WRITE,
    ),
    "resource.unmanage": _Cmd(
        cmd=resource.unmanage,
        required_permission=p.WRITE,
    ),
    "resource.unmove_unban": _Cmd(
        cmd=resource.unmove_unban,
        required_permission=p.WRITE,
    ),
    "sbd.disable_sbd": _Cmd(
        cmd=sbd.disable_sbd,
        required_permission=p.WRITE,
    ),
    "sbd.enable_sbd": _Cmd(
        cmd=sbd.enable_sbd,
        required_permission=p.WRITE,
    ),
    "scsi.unfence_node": _Cmd(
        cmd=scsi.unfence_node,
        required_permission=p.WRITE,
    ),
    "scsi.unfence_node_mpath": _Cmd(
        cmd=scsi.unfence_node_mpath,
        required_permission=p.WRITE,
    ),
    # deprecated, API v1 compatibility
    "status.full_cluster_status_plaintext": _Cmd(
        cmd=status.full_cluster_status_plaintext,
        required_permission=p.READ,
    ),
    # deprecated, API v1 compatibility
    "stonith_agent.describe_agent": _Cmd(
        cmd=stonith_agent.describe_agent,
        required_permission=p.READ,
    ),
    # deprecated, API v1 compatibility
    "stonith_agent.list_agents": _Cmd(
        cmd=stonith_agent.list_agents,
        required_permission=p.READ,
    ),
    "stonith.create": _Cmd(
        cmd=stonith.create,
        required_permission=p.WRITE,
    ),
    "stonith.create_in_group": _Cmd(
        cmd=stonith.create_in_group,
        required_permission=p.WRITE,
    ),
    # CMDs allowed in pcs_internal but not exposed via REST API:
    # "services.disable_service": Cmd(services.disable_service,
    # "services.enable_service": Cmd(services.enable_service,
    # "services.get_services_info": Cmd(services.get_services_info,
    # "services.start_service": Cmd(services.start_service,
    # "services.stop_service": Cmd(services.stop_service,
}


API_V1_COMPATIBILITY_MODE = (
    "resource_agent.describe_agent",
    "resource_agent.list_agents",
    "resource_agent.list_agents_for_standard_and_provider",
    "resource_agent.list_ocf_providers",
    "resource_agent.list_standards",
    "status.full_cluster_status_plaintext",
    "stonith_agent.describe_agent",
    "stonith_agent.list_agents",
)



import base64
import binascii
import logging
import pwd
import socket
import struct
from typing import (
    Optional,
    TypedDict,
    cast,
)

from tornado.http1connection import HTTP1Connection
from tornado.ioloop import IOLoop
from tornado.web import (
    HTTPError,
    RequestHandler,
)

from pcs.daemon.session import (
    Session,
    Storage,
)
from pcs.lib.auth.const import SUPERUSER
from pcs.lib.auth.provider import AuthProvider
from pcs.lib.auth.tools import (
    DesiredUser,
    get_effective_user,
)
from pcs.lib.auth.types import AuthUser

from .common import (
    LegacyApiBaseHandler,
    LegacyApiHandler,
    RoutesType,
)

PCSD_SESSION = "pcsd.sid"


class NotAuthorizedException(Exception):
    pass


class _BaseLibAuthProvider:
    def __init__(
        self, handler: RequestHandler, auth_provider: AuthProvider
    ) -> None:
        self._auth_provider = auth_provider
        self._handler = handler
        self._auth_logger = logging.getLogger("pcs.daemon.auth")

    async def login_user(self, username: str) -> AuthUser:
        auth_user = await IOLoop.current().run_in_executor(
            executor=None,
            func=lambda: self._auth_provider.login_user(username),
        )
        if auth_user is None:
            raise NotAuthorizedException()
        return auth_user

    def is_unix_socket_used(self) -> bool:
        # For whatever reason, handler.request.connection is typed as
        # HTTPConnection in tornado. That class, however, doesn't have stream
        # attribute. In reality, HTTP1Connection is probably used.
        return (
            cast(
                HTTP1Connection, self._handler.request.connection
            ).stream.socket.family
            == socket.AF_UNIX
        )

    def get_unix_socket_user(self) -> Optional[str]:
        if not self.is_unix_socket_used():
            return None

        # It is not cached to prevent inapropriate cache when handler is (in
        # hypotetical future) somehow reused. The responsibility for cache is
        # left to the place, that uses it.
        # For whatever reason, handler.request.connection is typed as
        # HTTPConnection in tornado. That class, however, doesn't have stream
        # attribute. In reality, HTTP1Connection is probably used.
        credentials = cast(
            HTTP1Connection, self._handler.request.connection
        ).stream.socket.getsockopt(
            socket.SOL_SOCKET,
            socket.SO_PEERCRED,
            struct.calcsize("3i"),
        )
        dummy_pid, uid, dummy_gid = struct.unpack("3i", credentials)
        if uid == 0:
            # treat root as cluster superuser
            return SUPERUSER

        return pwd.getpwuid(uid).pw_name

    async def auth_by_socket_user(self) -> AuthUser:
        username = self.get_unix_socket_user()
        if username:
            auth_user = await self.login_user(username)
            if auth_user:
                return auth_user
        raise NotAuthorizedException()


class PasswordAuthProvider(_BaseLibAuthProvider):
    async def auth_by_username_password(
        self, username: str, password: str
    ) -> AuthUser:
        auth_user = await IOLoop.current().run_in_executor(
            executor=None,
            func=lambda: self._auth_provider.auth_by_username_password(
                username, password
            ),
        )
        if auth_user is None:
            raise NotAuthorizedException()
        return auth_user


class TokenAuthProvider(_BaseLibAuthProvider):
    async def auth_by_token(self) -> AuthUser:
        token = self.auth_token
        if token:
            return await self._auth_by_token(token)
        return await self.auth_by_socket_user()

    async def _auth_by_token(self, token: str) -> AuthUser:
        auth_user = await IOLoop.current().run_in_executor(
            executor=None,
            func=lambda: self._auth_provider.auth_by_token(token),
        )
        if auth_user is None:
            raise NotAuthorizedException()
        return auth_user

    async def create_token(self, user: AuthUser) -> Optional[str]:
        return await IOLoop.current().run_in_executor(
            executor=None,
            func=lambda: self._auth_provider.create_token(user.username),
        )

    @property
    def auth_token(self) -> Optional[str]:
        return self._handler.get_cookie("token", default=None)


class LegacyTokenAuthProvider(TokenAuthProvider):
    async def auth_by_token_effective_user(self) -> tuple[AuthUser, AuthUser]:
        real_user = await self.auth_by_token()
        self._auth_logger.debug(
            "Real user=%s groups=%s",
            real_user.username,
            ",".join(real_user.groups),
        )
        if not real_user.is_superuser:
            return real_user, real_user
        effective_user = get_effective_user(
            real_user, self._get_effective_user()
        )
        self._auth_logger.debug(
            "Effective user=%s groups=%s",
            effective_user.username,
            ",".join(effective_user.groups),
        )
        return real_user, effective_user

    def _get_effective_user(self) -> DesiredUser:
        username = self._handler.get_cookie("CIB_user")
        groups = []
        if username:
            # use groups only if user is specified as well
            groups_raw = self._handler.get_cookie("CIB_user_groups")
            if groups_raw:
                try:
                    groups = (
                        base64.b64decode(groups_raw).decode("utf-8").split(" ")
                    )
                except (UnicodeError, binascii.Error):
                    self._auth_logger.warning("Unable to decode users groups")
        return DesiredUser(username, groups)


class SessionAuthProvider(_BaseLibAuthProvider):
    class CookieOptions(TypedDict):
        secure: bool
        httponly: bool
        samesite: str

    __cookie_options: CookieOptions = {
        "secure": True,
        "httponly": True,
        # rhbz#2097393
        # Prevent a cookie to be sent on cross-site requests, allow it to be
        # sent when navigating to pcs web UI from an external site.
        "samesite": "Lax",
    }

    def __init__(
        self,
        handler: RequestHandler,
        auth_provider: AuthProvider,
        session_storage: Storage,
    ) -> None:
        super().__init__(handler, auth_provider)
        self.__session_storage = session_storage
        self.__session: Optional[Session] = None

    def init_session(self) -> None:
        self.__session = self._get_session()

    async def auth_by_sid(self) -> AuthUser:
        if self.__session:
            return await self.login_user(self.__session.username)
        return await self.auth_by_socket_user()

    def _get_session(self) -> Optional[Session]:
        if self.__sid_from_client is None:
            return None
        return self.__session_storage.get(self.__sid_from_client)

    def session_logout(self) -> None:
        if self.__session is not None:
            self.__session_storage.destroy(self.__session.sid)
        self.__session = None
        self._sid_to_cookies()

    def _sid_to_cookies(self) -> None:
        """
        Write the session id into a response cookie.
        """
        if self.__session:
            self._handler.set_cookie(
                PCSD_SESSION, self.__session.sid, **self.__cookie_options
            )
        else:
            self._handler.clear_cookie(PCSD_SESSION)

    def put_request_cookies_sid_to_response_cookies_sid(self) -> None:
        """
        If sid came in the request cookies put it into response cookies. But do
        not start new one.
        """
        # TODO this method should exist temporarily (for sinatra compatibility)
        # pylint: disable=invalid-name
        if self.__sid_from_client is not None:
            self._handler.set_cookie(
                PCSD_SESSION, self.__sid_from_client, **self.__cookie_options
            )

    def is_sid_in_request_cookies(self) -> bool:
        return self.__sid_from_client is not None

    @property
    def __sid_from_client(self) -> Optional[str]:
        return self._handler.get_cookie(PCSD_SESSION, default=None)

    def update_session(self, auth_user: Optional[AuthUser]) -> None:
        if auth_user:
            if (
                not self.__session
                or self.__session.username != auth_user.username
            ):
                self.__session = self.__session_storage.login(
                    auth_user.username
                )
        else:
            self.__session = None
        self._sid_to_cookies()


class LegacyAuth(LegacyApiBaseHandler):
    _password_auth_provider: PasswordAuthProvider
    _token_auth_provider: TokenAuthProvider

    def initialize(self, auth_provider: AuthProvider) -> None:
        super().initialize()
        self._password_auth_provider = PasswordAuthProvider(self, auth_provider)
        self._token_auth_provider = TokenAuthProvider(self, auth_provider)

    async def auth(self) -> None:
        try:
            auth_user = (
                await self._password_auth_provider.auth_by_username_password(
                    self.get_body_argument("username") or "",
                    self.get_body_argument("password") or "",
                )
            )
            token = await self._token_auth_provider.create_token(auth_user)
            if token:
                self.write(token)
            else:
                raise HTTPError(400, reason="Unable to store token")
        except NotAuthorizedException:
            # To stay backward compatible with original ruby implementation,
            # an empty response needs to be returned if authentication fails
            pass

    async def post(self) -> None:
        await self.auth()

    async def get(self) -> None:
        await self.auth()


class LegacyTokenAuthenticationHandler(LegacyApiHandler):
    _token_auth_provider: LegacyTokenAuthProvider
    _auth_user: Optional[AuthUser]

    def initialize(self, auth_provider: AuthProvider) -> None:
        super().initialize()
        self._token_auth_provider = LegacyTokenAuthProvider(self, auth_provider)

    async def prepare(self) -> None:
        # pylint: disable=invalid-overridden-method
        super().prepare()
        try:
            (
                _,
                self._auth_user,
            ) = await self._token_auth_provider.auth_by_token_effective_user()
        except NotAuthorizedException as e:
            raise self.unauthorized() from e

    @property
    def auth_user(self) -> AuthUser:
        if not self._auth_user:
            raise self.unauthorized()
        return self._auth_user

    async def _handle_request(self) -> None:
        raise NotImplementedError()


class CheckAuth(LegacyTokenAuthenticationHandler):
    async def _handle_request(self) -> None:
        self.write('{"success":true}')


def get_routes(
    auth_provider: AuthProvider,
) -> RoutesType:
    auth_payload = dict(auth_provider=auth_provider)
    return [
        ("/remote/auth", LegacyAuth, auth_payload),
        ("/remote/check_auth", CheckAuth, auth_payload),
    ]



from lxml import etree

from pcs.common import (
    const,
    pacemaker,
    reports,
)
from pcs.common.reports.item import ReportItem
from pcs.lib import validate
from pcs.lib.cib.resource import group
from pcs.lib.cib.resource.common import get_parent_resource
from pcs.lib.cib.tools import (
    are_new_role_names_supported,
    find_unique_id,
    get_elements_by_ids,
)
from pcs.lib.errors import LibraryError
from pcs.lib.xml_tools import export_attributes

_BOOLEAN_VALUES = ("true", "false")

_ATTRIBUTES = ("action", "require-all", "role", "sequential")


def prepare_set(
    find_valid_id, resource_set, report_processor: reports.ReportProcessor
):
    """return resource_set with corrected ids"""
    if report_processor.report_list(
        _validate_options(resource_set["options"])
    ).has_errors:
        raise LibraryError()
    return {
        "ids": [find_valid_id(id) for id in resource_set["ids"]],
        "options": resource_set["options"],
    }


def _validate_options(options) -> reports.ReportItemList:
    # Pacemaker does not care currently about meaningfulness for concrete
    # constraint, so we use all attribs.
    validators = [
        validate.NamesIn(_ATTRIBUTES, option_type="set"),
        validate.ValueIn("action", const.PCMK_ACTIONS),
        validate.ValueIn("require-all", _BOOLEAN_VALUES),
        validate.ValueIn("role", const.PCMK_ROLES),
        validate.ValueIn("sequential", _BOOLEAN_VALUES),
        validate.ValueDeprecated(
            "role",
            {
                const.PCMK_ROLE_PROMOTED_LEGACY: const.PCMK_ROLE_PROMOTED,
                const.PCMK_ROLE_UNPROMOTED_LEGACY: const.PCMK_ROLE_UNPROMOTED,
            },
            reports.ReportItemSeverity.deprecation(),
        ),
    ]
    return validate.ValidatorAll(validators).validate(options)


def create(parent, resource_set):
    """
    parent - lxml element for append new resource_set
    """
    element = etree.SubElement(parent, "resource_set")
    if "role" in resource_set["options"]:
        resource_set["options"]["role"] = pacemaker.role.get_value_for_cib(
            resource_set["options"]["role"],
            is_latest_supported=are_new_role_names_supported(parent),
        )
    element.attrib.update(resource_set["options"])
    element.attrib["id"] = find_unique_id(
        parent.getroottree(),
        "{0}_set".format(parent.attrib.get("id", "constraint_set")),
    )

    for _id in resource_set["ids"]:
        etree.SubElement(element, "resource_ref").attrib["id"] = _id

    return element


def get_resource_id_set_list(element):
    return [
        resource_ref_element.attrib["id"]
        for resource_ref_element in element.findall(".//resource_ref")
    ]


def export(element):
    return {
        "ids": get_resource_id_set_list(element),
        "options": export_attributes(element),
    }


def is_resource_in_same_group(cib, resource_id_list):
    # We don't care about not found elements here, that is a job of another
    # validator. We do not care if the id doesn't belong to a resource either
    # for the same reason.
    element_list, _ = get_elements_by_ids(cib, set(resource_id_list))

    parent_list = []
    for element in element_list:
        parent = get_parent_resource(element)
        if parent is not None and group.is_group(parent):
            parent_list.append(parent)

    if len(set(parent_list)) != len(parent_list):
        raise LibraryError(
            ReportItem.error(
                reports.messages.CannotSetOrderConstraintsForResourcesInTheSameGroup()
            )
        )
